# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import logging

from .util import LoggerWrapper, GetIndex, GetRNG, get_cls_params
from .value import ParseValue, GetAllParams, CheckAllParams, SetDefaultIndex
from .input import RegisterInputConnectedType
from ..errors import GalSimConfigError, GalSimConfigValueError
from ..wcs import BaseWCS, PixelScale, OffsetWCS, JacobianWCS, AffineTransform
from ..wcs import ShearWCS, OffsetShearWCS, UVFunction, RaDecFunction
from ..fitswcs import TanWCS, FitsWCS
from ..angle import Angle, AngleUnit
from ..position import PositionD
from ..celestial import CelestialCoord
from ..utilities import basestring

# This file handles the construction of wcs types in config['image']['wcs'].

# This module-level dict will store all the registered wcs types.
# See the RegisterWCSType function at the end of this file.
# The keys are the (string) names of the wcs types, and the values will be builders that know
# how to build the WCS object.
valid_wcs_types = {}


def BuildWCS(config, key, base, logger=None):
    """Read the wcs parameters from config[key] and return a constructed wcs object.

    Parameters:
        config:     A dict with the configuration information. (usually base['image'])
        key:        The key name in config indicating which object to build.
        base:       The base dict of the configuration.
        logger:     Optionally, provide a logger for logging debug statements. [default: None]

    Returns:
        a BaseWCS instance
    """
    logger = LoggerWrapper(logger)
    logger.debug('image %d: Start BuildWCS key = %s',base.get('image_num',0),key)

    try:
        param = config[key]
    except KeyError:
        # Default if no wcs is to use PixelScale
        if 'pixel_scale' in config:
            scale = ParseValue(config, 'pixel_scale', base, float)[0]
        else:
            scale = 1.0
        return PixelScale(scale)

    # Check for direct value, else get the wcs type
    if isinstance(param, BaseWCS):
        return param
    elif isinstance(param, basestring) and (param[0] == '$' or param[0] == '@'):
        return ParseValue(config, key, base, None)[0]
    elif isinstance(param, dict):
        wcs_type = param.get('type', 'PixelScale')
    else:
        raise GalSimConfigError("wcs must be either a BaseWCS or a dict")

    # For these two, just do the usual ParseValue function.
    if wcs_type in ('Eval', 'Current'):
        return ParseValue(config, key, base, None)[0]

    # Check if we can use the current cached object
    orig_index_key = base.get('index_key','image_num')
    base['index_key'] = 'image_num'
    index, index_key = GetIndex(param, base)
    if 'current' in param:
        cwcs, csafe, cvalue_type, cindex, cindex_key = param['current']
        if cindex == index:
            logger.debug('image %d: The wcs object is already current', base.get('image_num',0))
            logger.debug('image %d: index_key = %s, index = %d',base.get('image_num',0),
                         cindex_key, cindex)
            return cwcs

    # Special case: origin == center means to use image_center for the wcs origin
    if 'origin' in param and param['origin'] == 'center':
        origin = base['image_center']
        param['origin'] = origin

    if wcs_type not in valid_wcs_types:
        raise GalSimConfigValueError("Invalid image.wcs.type.", wcs_type,
                                     list(valid_wcs_types.keys()))
    logger.debug('image %d: Building wcs type %s', base.get('image_num',0), wcs_type)
    builder = valid_wcs_types[wcs_type]
    wcs = builder.buildWCS(param, base, logger)
    logger.debug('image %d: wcs = %s', base.get('image_num',0), wcs)

    param['current'] = wcs, False, None, index, index_key
    base['index_key'] = orig_index_key

    return wcs


class WCSBuilder:
    """A base class for building WCS objects.

    The base class defines the call signatures of the methods that any derived class should follow.
    """
    def buildWCS(self, config, base, logger):
        """Build the WCS based on the specifications in the config dict.

        Note: Sub-classes must override this function with a real implementation.

        Parameters:
            config:     The configuration dict for the wcs type.
            base:       The base configuration dict.
            logger:     If provided, a logger for logging debug statements.

        Returns:
            the constructed WCS object.
        """
        raise NotImplementedError("The %s class has not overridden buildWCS"%self.__class__)


class SimpleWCSBuilder(WCSBuilder):
    """A class for building simple WCS objects.

    The initializer takes an init_func, which is the class or function to call to build the WCS.
    For the kwargs, it calls getKwargs, which does the normal parsing of the req_params and
    related class attributes.
    """
    def __init__(self, init_func):
        self.init_func = init_func

    def getKwargs(self, build_func, config, base):
        """Get the kwargs to pass to the build function based on the following attributes of
        build_func:

        _req_params
                        A dict of required parameters and their types.
        _opt_params
                        A dict of optional parameters and their types.
        _single_params
                        A list of dicts of parameters such that one and only one of
                        parameter in each dict is required.
        _takes_rng
                        A bool value saying whether an rng object is required.
                        (Which would be weird for this, but it's part of our standard set.)

        See any of the classes in wcs.py for examples of classes that set these attributes.

        Parameters:
            build_func: The class or function from which to get the
            config:     The configuration dict for the wcs type.
            base:       The base configuration dict.

        Returns:
            kwargs
        """
        # Then use the standard trick of reading the required and optional parameters
        # from the class or function attributes.
        req, opt, single, takes_rng = get_cls_params(build_func)

        # Pull in the image layer pixel_scale as a scale item if necessary.
        if ( ('scale' in req or 'scale' in opt) and 'scale' not in config and
            'pixel_scale' in base['image'] ):
            config['scale'] = base['image']['pixel_scale']

        kwargs, safe = GetAllParams(config, base, req, opt, single)

        # This would be weird, but might as well check...
        if takes_rng: # pragma: no cover
            kwargs['rng'] = GetRNG(config, base)
        return kwargs

    def buildWCS(self, config, base, logger):
        """Build the WCS based on the specifications in the config dict.

        Parameters:
            config:     The configuration dict for the wcs type.
            base:       The base configuration dict.
            logger:     If provided, a logger for logging debug statements.

        Returns:
            the constructed WCS object.
        """
        kwargs = self.getKwargs(self.init_func,config,base)
        return self.init_func(**kwargs)


class OriginWCSBuilder(SimpleWCSBuilder):
    """A specialization for WCS classes that use a different type depending on whether there
    is an origin or world_origin parameter in the config dict.
    """
    def __init__(self, init_func, origin_init_func):
        self.init_func = init_func
        self.origin_init_func = origin_init_func

    def buildWCS(self, config, base, logger):
        """Build the WCS based on the specifications in the config dict, using the appropriate
        type depending on whether an origin is provided.

        Parameters:
            config:     The configuration dict for the wcs type.
            base:       The base configuration dict.
            logger:     If provided, a logger for logging debug statements.

        Returns:
            the constructed WCS object.
        """
        if 'origin' in config or 'world_origin' in config:
            build_func = self.origin_init_func
        else:
            build_func = self.init_func
        kwargs = self.getKwargs(build_func,config,base)
        return build_func(**kwargs)


class TanWCSBuilder(WCSBuilder):
    """The TanWCS type needs special handling to get the kwargs, since the TanWCS function
    takes an AffineTransform as one of the arguments, so we need to build that from
    dudx, dudy, etc.  We also need to construct a CelestialCoord object for the world_origin,
    which we make from ra, dec paramters.
    """
    def __init__(self): pass

    def buildWCS(self, config, base, logger):
        """Build the TanWCS based on the specifications in the config dict.

        Parameters:
            config:     The configuration dict for the wcs type.
            base:       The base configuration dict.
            logger:     If provided, a logger for logging debug statements.

        Returns:
            the constructed WCS object.
        """
        req = { "dudx" : float, "dudy" : float, "dvdx" : float, "dvdy" : float,
                "ra" : Angle, "dec" : Angle }
        opt = { "units" : str, "origin" : PositionD }
        params, safe = GetAllParams(config, base, req=req, opt=opt)

        dudx = params['dudx']
        dudy = params['dudy']
        dvdx = params['dvdx']
        dvdy = params['dvdy']
        ra = params['ra']
        dec = params['dec']
        units = params.get('units', 'arcsec')
        origin = params.get('origin', None)

        affine = AffineTransform(dudx, dudy, dvdx, dvdy, origin)
        world_origin = CelestialCoord(ra, dec)
        units = AngleUnit.from_name(units)

        return TanWCS(affine=affine, world_origin=world_origin, units=units)

class ListWCSBuilder(WCSBuilder):
    """Select a wcs from a list
    """
    def buildWCS(self, config, base, logger):
        req = { 'items' : list }
        opt = { 'index' : int }
        # Only Check, not Get.  We need to handle items a bit differently, since it's a list.
        CheckAllParams(config, req=req, opt=opt)
        items = config['items']
        if not isinstance(items,list):
            raise GalSimConfigError("items entry for type=List is not a list.")

        # Setup the indexing sequence if it hasn't been specified using the length of items.
        SetDefaultIndex(config, len(items))
        index, safe = ParseValue(config, 'index', base, int)

        if index < 0 or index >= len(items):
            raise GalSimConfigError("index %d out of bounds for wcs type=List"%index)
        return BuildWCS(items, index, base)

def RegisterWCSType(wcs_type, builder, input_type=None):
    """Register a wcs type for use by the config apparatus.

    Parameters:
        wcs_type:       The name of the type in config['image']['wcs']
        builder:        A builder object to use for building the WCS object.  It should
                        be an instance of a subclass of WCSBuilder.
        input_type:     If the WCS builder utilises an input object, give the key name of the
                        input type here.  (If it uses more than one, this may be a list.)
                        [default: None]
    """
    valid_wcs_types[wcs_type] = builder
    RegisterInputConnectedType(input_type, wcs_type)


RegisterWCSType('PixelScale', OriginWCSBuilder(PixelScale, OffsetWCS))
RegisterWCSType('Shear', OriginWCSBuilder(ShearWCS, OffsetShearWCS))
RegisterWCSType('Jacobian', OriginWCSBuilder(JacobianWCS, AffineTransform))
RegisterWCSType('Affine', OriginWCSBuilder(JacobianWCS, AffineTransform))
RegisterWCSType('UVFunction', SimpleWCSBuilder(UVFunction))
RegisterWCSType('RaDecFunction', SimpleWCSBuilder(RaDecFunction))
RegisterWCSType('Fits', SimpleWCSBuilder(FitsWCS))
RegisterWCSType('Tan', TanWCSBuilder())
RegisterWCSType('List', ListWCSBuilder())



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'GalaxySample', 'COSMOSCatalog', ]

import numpy as np
import math
import os

from .real import RealGalaxy, RealGalaxyCatalog, _parse_files_dirs
from .errors import GalSimError, GalSimValueError, GalSimIncompatibleValuesError
from .errors import GalSimNotImplementedError, galsim_warn
from .utilities import lazy_property
from ._pyfits import pyfits
from .random import BaseDeviate
from . import utilities
from .bandpass import Bandpass
from .sed import SED
from .angle import radians
from .exponential import Exponential
from .sersic import DeVaucouleurs, Sersic

# Below is a number that is needed to relate the COSMOS parametric galaxy fits to quantities that
# GalSim needs to make a GSObject representing that fit.  It is simply the pixel scale, in arcsec,
# in the COSMOS weak lensing reductions used for the fits.
# Note: This isn't used anywhere.  This is just informational, really.
cosmos_pix_scale = 0.03

class GalaxySample:
    """
    A class representing a random subsample of galaxies from an arbitrary deep survey.

    Depending on the keyword arguments, particularly ``use_real``, a `GalaxySample` may be able to
    generate real galaxies, parametric galaxies, or both.

    The original version of this functionality is now in the subclass `COSMOSCatalog`, which
    specializes to HST observations of the COSMOS field.  `GalaxySample` is a generalization of
    that, which works for arbitrary data sets, although the user is responsible for building
    the appropriate input files to use with it.

    Unlike `COSMOSCatalog`, which has easy options for picking out one of the two galaxy subsets
    that are available for download using ``galsim_download_cosmos``, for this class you need to
    manually specify the file name.

        >>> sample = galsim.GalaxySample(file_name)

    Other than this difference, the functionality of this class is the same as `COSMOSCatalog`.
    See the documentation of that function for more detail.

    Parameters:
        file_name:          The file containing the catalog.
        dir:                The directory with the catalog file and, if making realistic galaxies,
                            the image and noise files (or symlinks to them). [default: None, which
                            will look in $PREFIX/share/galsim.]
        preload:            Keyword that is only used for real galaxies, not parametric ones, to
                            choose whether to preload the header information.  If ``preload=True``,
                            the bulk of the I/O time is in the constructor.  If ``preload=False``,
                            there is approximately the same total I/O time (assuming you eventually
                            use most of the image files referenced in the catalog), but it is spread
                            over the calls to makeGalaxy().  [default: False]
        orig_exptime:       The exposure time (in seconds) of the original observations.
                            [default: 1]
        orig_area:          The effective collecting area (in cm^2) of the original observations.
                            [default: 1]
        use_real:           Enable the use of realistic galaxies?  [default: True]
                            If this parameter is False, then ``makeGalaxy(gal_type='real')`` will
                            not be allowed, and there will be a (modest) decrease in RAM and time
                            spent on I/O when initializing the COSMOSCatalog. If the real
                            catalog is not available for some reason, it will still be possible to
                            make parametric images.
        exclusion_level:    Level of additional cuts to make on the galaxies based on the quality
                            of postage stamp definition and/or parametric fit quality [beyond the
                            minimal cuts imposed when making the catalog - see Mandelbaum et
                            al. (2012, MNRAS, 420, 1518) for details].  Options:

                            - "none": No cuts.
                            - "bad_stamp": Apply cuts to eliminate galaxies that have failures in
                              postage stamp definition.  These cuts may also eliminate a small
                              subset of the good postage stamps as well.
                            - "bad_fits": Apply cuts to eliminate galaxies that have failures in the
                              parametric fits.  These cuts may also eliminate a small
                              subset of the good parametric fits as well.
                            - "marginal": Apply the above cuts, plus ones that eliminate some more
                              marginal cases.

                            Use of "bad_stamp" or "marginal" requires a ``CATALOG_selection.fits``
                            file (where CATALOG is ``file_name`` without the ".fits" extension).
                            [default: "none"]
        min_hlr:            Exclude galaxies whose fitted half-light radius is smaller than this
                            value (in arcsec).  [default: 0, meaning no limit]
        max_hlr:            Exclude galaxies whose fitted half-light radius is larger than this
                            value (in arcsec).  [default: 0, meaning no limit]
        min_flux:           Exclude galaxies whose fitted flux is smaller than this value.
                            [default: 0, meaning no limit]
        max_flux:           Exclude galaxies whose fitted flux is larger than this value.
                            [default: 0, meaning no limit]
        cut_ratio:          For the "bad_stamp" exclusions, cut out any stamps with average
                            adjacent pixels larger than this fraction of the peak pixel count.
                            [default: 0.8]
        sn_limit:           For the "bad_stamp" exclusions, cut out any stamps with estimated
                            S/N for an elliptical Gaussian less than this limit. [default: 10.]
        min_mask_dist:      For the "bad_stamp" exclusions, remove any stamps that have some
                            masked pixels closer to the center than this minimum distance
                            (in pixels). [default: 10]
        exptime:            The exposure time (in seconds) to assume when creating galaxies.
                            [default: None, which means to use orig_exptime]
        area:               The effective collecting area (in cm^2) to assume when creating
                            galaxies. [default: None, which means to use orig_area]

    After construction, the following attributes are available:

    Attributes:
        nobjects:       The number of objects in the sample
    """
    _opt_params = { 'file_name' : str, 'dir' : str,
                    'orig_exptime': float, 'orig_area': float,
                    'preload' : bool, 'use_real' : bool,
                    'exclusion_level' : str, 'min_hlr' : float, 'max_hlr' : float,
                    'min_flux' : float, 'max_flux' : float,
                    'cut_ratio' : float, 'sn_limit' : float, 'min_mask_dist': float,
                  }

    def __init__(self, file_name, dir=None, preload=False,
                 orig_exptime=1., orig_area=1.,
                 use_real=True, exclusion_level='marginal', min_hlr=0, max_hlr=0.,
                 min_flux=0., max_flux=0., cut_ratio=0.8, sn_limit=10., min_mask_dist=11,
                 exptime=None, area=None, _use_sample=None):

        self.use_real = use_real
        self.preload = preload
        self.cut_ratio = cut_ratio
        self.sn_limit = sn_limit
        self.min_mask_dist = min_mask_dist
        self.orig_exptime = orig_exptime
        self.orig_area = orig_area
        self.exptime = exptime
        self.area = area

        # We'll set these up if and when we need them.
        self._bandpass = None
        self._sed = None

        if exclusion_level not in ('none', 'bad_stamp', 'bad_fits', 'marginal'):
            raise GalSimValueError("Invalid value of exclusion_level.", exclusion_level,
                                   ('none', 'bad_stamp', 'bad_fits', 'marginal'))

        # Parse the file name
        self.full_file_name, _, _ = _parse_files_dirs(file_name, dir, None)
        self.use_sample = _use_sample

        try:
            # Read in data.
            with pyfits.open(self.full_file_name) as fits:
                self.param_cat = fits[1].data
            # Check if this was the right file.  It should have a 'fit_status' column.
            self.param_cat['fit_status']
        except KeyError:
            # But if that doesn't work, then the name might be the name of the real catalog,
            # so try adding _fits to it as above.
            param_file_name = self.full_file_name.replace('.fits', '_fits.fits')
            with pyfits.open(param_file_name) as fits:
                self.param_cat = fits[1].data

        # NB. The pyfits FITS_Rec class has a bug where it makes a copy of the full
        # record array in each record (e.g. in getParametricRecord) and then doesn't
        # garbage collect it until the top-level FITS_Record goes out of scope.
        # This leads to a memory leak of order 10MB or so each time we make a parametric
        # galaxy.
        # cf. https://mail.scipy.org/pipermail/astropy/2014-June/003218.html
        # also https://github.com/astropy/astropy/pull/520
        # The simplest workaround seems to be to convert it to a regular numpy recarray.
        # (This also makes it run much faster, as an extra bonus!)
        self.param_cat = np.array(self.param_cat, copy=True)

        self.orig_index = np.arange(len(self.param_cat))
        self._apply_exclusion(exclusion_level, min_hlr, max_hlr, min_flux, max_flux)

    @lazy_property
    def real_cat(self):
        if self.use_real:
            return RealGalaxyCatalog(self.full_file_name, preload=self.preload)
        else:
            return None

    def _apply_exclusion(self, exclusion_level, min_hlr=0, max_hlr=0, min_flux=0, max_flux=0):
        mask = np.ones(len(self.orig_index), dtype=bool)
        if exclusion_level in ('marginal', 'bad_stamp'):
            # First, read in what we need to impose selection criteria, if the appropriate
            # exclusion_level was chosen.

            # This should work if the user passed in (or we defaulted to) the real galaxy
            # catalog name:
            selection_file_name = self.full_file_name.replace('.fits', '_selection.fits')
            try:
                with pyfits.open(selection_file_name) as fits:
                    self.selection_cat = fits[1].data
            except (OSError):
                # There's one more option: full_file_name might be the parametric fit file, so
                # we have to strip off the _fits.fits (instead of just the .fits)
                selection_file_name = self.full_file_name.replace('_fits', '_selection')
                try:
                    with pyfits.open(selection_file_name) as fits:
                        self.selection_cat = fits[1].data
                except (OSError):  # pragma: no cover
                    if self.use_sample is None:
                        raise
                    else:
                        raise OSError("File with GalSim selection criteria not found. "
                                      "Run the program `galsim_download_cosmos -s %s` to get the "
                                      "necessary selection file."%(self.use_sample))

            # We proceed to select galaxies in a way that excludes suspect postage stamps (e.g.,
            # with deblending issues), suspect parametric model fits, or both of the above plus
            # marginal ones.  These two options for 'exclusion_level' involve placing cuts on
            # the S/N of the object detection in the original postage stamp, and on issues with
            # masking that can indicate deblending or detection failures.  These cuts were used
            # in GREAT3.  In the case of the masking cut, in some cases there are messed up ones
            # that have a 0 for self.selection_cat['peak_image_pixel_count'].  To make sure we
            # don't divide by zero (generating a RuntimeWarning), and still eliminate those, we
            # will first set that column to 1.e-5.  We choose a sample-dependent mask ratio cut,
            # since this depends on the peak object flux, which will differ for the two samples
            # (and we can't really cut on this for arbitrary user-defined samples).
            div_val = self.selection_cat['peak_image_pixel_count']
            div_val[div_val == 0.] = 1.e-5
            mask &= ( (self.selection_cat['sn_ellip_gauss'] >= self.sn_limit) &
                      ((self.selection_cat['min_mask_dist_pixels'] > self.min_mask_dist) |
                       (self.selection_cat['average_mask_adjacent_pixel_count'] / \
                           div_val < self.cut_ratio)) )

            # Finally, impose a cut that the total flux in the postage stamp should be positive,
            # which excludes a tiny number of galaxies (of order 10 in each sample) with some sky
            # subtraction or deblending errors.  Some of these are eliminated by other cuts when
            # using exclusion_level='marginal'.
            if self.real_cat is not None:
                mask &= self.real_cat.stamp_flux > 0

        if exclusion_level in ('bad_fits', 'marginal'):
            # This 'exclusion_level' involves eliminating failed parametric fits (bad fit status
            # flags).  In this case we only get rid of those with failed bulge+disk AND failed
            # Sersic fits, so there is no viable parametric model for the galaxy.
            sersicfit_status = self.param_cat['fit_status'][:,4]
            bulgefit_status = self.param_cat['fit_status'][:,0]
            mask &= ( ((sersicfit_status > 0) &
                      (sersicfit_status < 5)) |
                      ((bulgefit_status > 0) &
                      (bulgefit_status < 5)) )

        if exclusion_level == 'marginal':
            # We have already placed some cuts (above) in this case, but we'll do some more.  For
            # example, a failed bulge+disk fit often indicates difficulty in fit convergence due to
            # noisy surface brightness profiles, so we might want to toss out those that have a
            # failure in EITHER fit.
            mask &= ( ((sersicfit_status > 0) &
                      (sersicfit_status < 5)) &
                      ((bulgefit_status > 0) &
                      (bulgefit_status < 5)) )

            # Some fit parameters can indicate a likely sky subtraction error: very high sersic n
            # AND abnormally large half-light radius (>1 arcsec).
            if 'hlr' not in self.param_cat.dtype.names and self.use_sample is not None:  # pragma: no cover
                raise OSError("You still have the old COSMOS catalog.  Run the program "
                              "`galsim_download_cosmos -s %s` to upgrade."%(self.use_sample))
            hlr = self.param_cat['hlr'][:,0]
            n = self.param_cat['sersicfit'][:,2]
            mask &= ( (n < 5) | (hlr < 1.) )

            # Major flux differences in the parametric model vs. the COSMOS catalog can indicate fit
            # issues, deblending problems, etc.
            mask &= ( np.abs(self.selection_cat['dmag']) < 0.8)

        if min_hlr > 0. or max_hlr > 0. or min_flux > 0. or max_flux > 0.:
            if 'hlr' not in self.param_cat.dtype.names and self.use_sample is not None:  # pragma: no cover
                raise OSError("You still have the old COSMOS catalog.  Run the program "
                              "`galsim_download_cosmos -s %s` to upgrade."%(self.use_sample))

            hlr = self.param_cat['hlr'][:,0] # sersic half-light radius
            flux = self.param_cat['flux'][:,0]

            if min_hlr > 0.:
                mask &= (hlr > min_hlr)
            if max_hlr > 0.:
                mask &= (hlr < max_hlr)
            if min_flux > 0.:
                mask &= (flux > min_flux)
            if max_flux > 0.:
                mask &= (flux < max_flux)

        self.orig_index = self.orig_index[mask]
        self.nobjects = len(self.orig_index)

    # We need this method because the config apparatus will use this via a Proxy, and they cannot
    # access attributes directly -- just call methods.  So this is how we get nobjects there.
    def getNObjects(self) : return self.nobjects
    def getUseSample(self): return self.use_sample
    def getOrigIndex(self, index): return self.orig_index[index]
    def getNTot(self) : return len(self.param_cat)
    def __len__(self): return self.nobjects

    def makeGalaxy(self, index=None, gal_type=None, chromatic=False, noise_pad_size=5,
                   deep=False, sersic_prec=0.05, rng=None, n_random=None, gsparams=None):
        """
        Routine to construct one or more `GSObject` instances corresponding to the catalog entry
        with a particular index or indices.

        The flux of the galaxy corresponds to a 1 second exposure time with the Hubble Space
        Telescope.  Users who wish to simulate F814W images with a different telescope and an
        exposure time longer than 1 second should multiply by that exposure time, and by the square
        of the ratio of the effective diameter of their telescope compared to that of HST.
        (Effective diameter may differ from the actual diameter if there is significant
        obscuration.)  See demo11.py for an example that explicitly takes this normalization into
        account.

        Due to the adopted flux normalization, drawing into an image with the COSMOS bandpass,
        zeropoint of 25.94, and pixel scale should give the right pixel values to mimic the actual
        COSMOS science images.  The COSMOS science images that we use are normalized to a count rate
        of 1 second, which is why there is no need to rescale to account for the COSMOS exposure
        time.

        There is an option to make chromatic objects (``chromatic=True``); however, it is important
        to bear in mind that we do not actually have spatially-resolved color information for these
        galaxies, so this keyword can only be True if we are using parametric galaxies.  Even then,
        we simply do the most arbitrary thing possible, which is to assign bulges an elliptical
        `SED`, disks a disk-like `SED`, and `Sersic` galaxies with intermediate values of n some
        intermediate `SED`.  We assume that the photometric redshift is the correct redshift for
        these galaxies (which is a good assumption for COSMOS 30-band photo-z for these bright
        galaxies).  For the given `SED` and redshift, we then normalize to give the right (observed)
        flux in F814W.  Note that for a mock "deep" sample, the redshift distributions of the
        galaxies would be modified, which is not included here.

        For this chromatic option, it is still the case that the output flux normalization is
        appropriate for the HST effective telescope diameter and a 1 second exposure time, so users
        who are simulating another scenario should account for this.

        Note that the returned objects use arcsec for the units of their linear dimension.  If you
        are using a different unit for other things (the PSF, WCS, etc.), then you should dilate
        the resulting object with ``gal.dilate(galsim.arcsec / scale_unit)``.

        Parameters:
            index:          Index of the desired galaxy in the catalog for which a `GSObject`
                            should be constructed.  You may also provide a list or array of
                            indices, in which case a list of objects is returned. If None,
                            then a random galaxy (or more: see n_random kwarg) is chosen,
                            correcting for catalog-level selection effects if weights are
                            available. [default: None]
            gal_type:       Either 'real' or 'parametric'.  This determines which kind of
                            galaxy model is made. [If catalog was loaded with ``use_real=False``,
                            then this defaults to 'parametric', and in fact 'real' is
                            not allowed.  If catalog was loaded with ``use_real=True``, then
                            this defaults to 'real'.]
            chromatic:      Make this a chromatic object, or not?  [default: False]
            noise_pad_size: For realistic galaxies, the size of region to pad with noise,
                            in arcsec.  [default: 5, an arbitrary, but not completely
                            ridiculous choice.]
            deep:           Modify fluxes and sizes of galaxies from the F814W<23.5 sample in
                            order to roughly simulate an F814W<25 sample but with higher S/N, as
                            in GREAT3? [default: False]  Note that this keyword will be ignored
                            (except for issuing a warning) if the input catalog already
                            represents the F814W<25.2 sample.
            sersic_prec:    The desired precision on the Sersic index n in parametric galaxies.
                            GalSim is significantly faster if it gets a smallish number of
                            Sersic values, so it can cache some of the calculations and use
                            them again the next time it gets a galaxy with the same index.
                            If ``sersic_prec`` is 0.0, then use the exact value of index n from
                            the catalog.  But if it is >0, then round the index to that
                            precision.  [default: 0.05]
            rng:            A random number generator to use for selecting a random galaxy
                            (may be any kind of `BaseDeviate` or None) and to use in generating
                            any noise field when padding.  [default: None]
            n_random:       The number of random galaxies to build, if 'index' is None.
                            [default: 1]
            gsparams:       An optional `GSParams` argument. [default: None]

        Returns:
            Either a `GSObject` or a `ChromaticObject` depending on the value of ``chromatic``,
            or a list of them if ``index`` is an iterable.
        """
        return self._makeGalaxy(self, index, gal_type, chromatic, noise_pad_size,
                                deep, sersic_prec, self.exptime, self.area,
                                rng, n_random, gsparams)

    @staticmethod
    def _makeGalaxy(self, index=None, gal_type=None, chromatic=False, noise_pad_size=5,
                    deep=False, sersic_prec=0.05, exptime=None, area=None,
                    rng=None, n_random=None, gsparams=None):
        if not self.canMakeReal():
            if gal_type is None:
                gal_type = 'parametric'
            elif gal_type != 'parametric':
                raise GalSimIncompatibleValuesError(
                    "Only 'parametric' galaxy type is allowed when use_real == False",
                    gal_type=gal_type, use_real=self.canMakeReal())
        else:
            if gal_type is None:
                gal_type = 'real'

        if gal_type not in ('real', 'parametric'):
            raise GalSimValueError("Invalid galaxy type %r", gal_type, ('real', 'parametric'))

        # Make rng if we will need it.
        if index is None or gal_type == 'real':
            if rng is None:
                rng = BaseDeviate()
            elif not isinstance(rng, BaseDeviate):
                raise TypeError("The rng provided to makeGalaxy is not a BaseDeviate")

        # Select random indices if necessary (no index given).
        if index is None:
            if n_random is None: n_random = 1
            index = self.selectRandomIndex(n_random, rng=rng)
        else:
            if n_random is not None:
                raise GalSimIncompatibleValuesError(
                    "Cannot specify both index and n_random", n_random=n_random, index=index)

        if hasattr(index, '__iter__'):
            indices = index
        else:
            indices = [index]

        # Check whether this is a COSMOSCatalog meant to represent real or parametric objects, then
        # call the appropriate helper routine for that case.
        if gal_type == 'real':
            if chromatic:
                raise GalSimNotImplementedError("Cannot yet make real chromatic galaxies!")
            gal_list = []
            for idx in indices:
                real_params = self.getRealParams(idx)
                gal = RealGalaxy(real_params, noise_pad_size=noise_pad_size, rng=rng,
                                 gsparams=gsparams)
                gal_list.append(gal)

        else:
            if chromatic:
                bandpass = self.getBandpass()
                sed = self.getSED()
            else:
                bandpass = None
                sed = None
            gal_list = []
            for idx in indices:
                record = self.getParametricRecord(idx)
                gal = COSMOSCatalog._buildParametric(record, sersic_prec, gsparams,
                                                     chromatic, bandpass, sed)
                gal_list.append(gal)

        flux_scaling = 1.
        if exptime is not None:
            flux_scaling *= exptime / self.orig_exptime
        if area is not None:
            flux_scaling *= area / self.orig_area
        if flux_scaling != 1.:
            gal_list = [gal * flux_scaling for gal in gal_list]

        # If trying to use the 23.5 sample and "fake" a deep sample, rescale the size and flux as
        # suggested in the GREAT3 handbook.
        if deep:
            if self.getUseSample() == '23.5':
                # Rescale the flux to get a limiting mag of 25 in F814W when starting with a
                # limiting mag of 23.5.  Make the galaxies a factor of 0.6 smaller and appropriately
                # fainter.
                flux_factor = 10.**(-0.4*1.5)
                size_factor = 0.6
                gal_list = [ gal.dilate(size_factor) * flux_factor for gal in gal_list ]
            elif self.getUseSample() == '25.2':
                galsim_warn("Ignoring `deep` argument, because the sample being used already "
                            "corresponds to a flux limit of F814W<25.2")
            else:
                galsim_warn("Ignoring `deep` argument, because the sample being used does not "
                            "corresponds to a flux limit of F814W<23.5")

        # Store the orig_index as gal.index regardless of whether we have a RealGalaxy or not.
        # It gets set as part of making a real galaxy, but not by _buildParametric.
        # And if we are doing the deep scaling, then it gets messed up by that.
        # So just put it in here at the end to be sure.
        for gal, idx in zip(gal_list, indices):
            gal.index = self.getOrigIndex(idx)
            if hasattr(gal, 'original'): gal.original.index = gal.index

        if hasattr(index, '__iter__'):
            return gal_list
        else:
            return gal_list[0]

    def selectRandomIndex(self, n_random=1, rng=None, _n_rng_calls=False):
        """
        Routine to select random indices out of the catalog.  This routine does a weighted random
        selection with replacement (i.e., there is no guarantee of uniqueness of the selected
        indices).  Weighting uses the weight factors available in the catalog, if any; these weights
        are typically meant to remove any selection effects in the catalog creation process.

        Parameters:
            n_random:   Number of random indices to return. [default: 1]
            rng:        A random number generator to use for selecting a random galaxy
                        (may be any kind of `BaseDeviate` or None). [default: None]

        Returns:
            A single index if n_random==1 or a NumPy array containing the randomly-selected
            indices if n_random>1.
        """
        # Set up the random number generator.
        if rng is None:
            rng = BaseDeviate()

        if self.real_cat is not None:
            use_weights = self.real_cat.weight[self.orig_index]
        else:
            galsim_warn("Selecting random object without correcting for catalog-level "
                        "selection effects.  This correction requires the existence of "
                        "real catalog with valid weights in addition to parametric one. "
                        "Create the COSMOSCatalog with use_real=True to avoid this warning.")
            use_weights = None

        # By default, get the number of RNG calls.  We then decide whether or not to return them
        # based on _n_rng_calls.
        index, n_rng_calls = utilities.rand_with_replacement(
                n_random, self.nobjects, rng, use_weights, _n_rng_calls=True)

        if n_random>1:
            if _n_rng_calls:
                return index, n_rng_calls
            else:
                return index
        else:
            if _n_rng_calls:
                return index[0], n_rng_calls
            else:
                return index[0]

    def getBandpass(self):
        # Defer making the Bandpass and reading in SEDs until we actually are going to use them.
        # It's not a huge calculation, but the thin() call especially isn't trivial.

        if self._bandpass is None:
            # We have to set an appropriate zeropoint.  This is slightly complicated: The
            # nominal COSMOS zeropoint for single-orbit depth (2000s of usable exposure time,
            # across 4 dithered exposures) is supposedly 25.94.  But the science images that we
            # are using were normalized to count rate, not counts, meaning that an object with
            # mag=25.94 has a count rate of 1 photon/sec, not 1 photon total.  Since we've
            # declared our flux normalization for the outputs to be appropriate for a 1s
            # exposure, we use this zeropoint directly.
            # This means that when drawing chromatic parametric galaxies, the outputs will be
            # properly normalized in terms of counts.
            zp = 25.94
            self._bandpass = Bandpass('ACS_wfc_F814W.dat', wave_type='nm').withZeropoint(zp)
        return self._bandpass

    def getSED(self):
        if self._sed is None:
            # Read in some SEDs.  We are using some fairly truncated and thinned ones, because
            # in any case the SED assignment here is somewhat arbitrary and should not be taken
            # too seriously.
            self._sed = [
                # bulge
                SED('CWW_E_ext_more.sed', wave_type='Ang', flux_type='flambda'),
                # disk
                SED('CWW_Scd_ext_more.sed', wave_type='Ang', flux_type='flambda'),
                # intermediate
                SED('CWW_Sbc_ext_more.sed', wave_type='Ang', flux_type='flambda')
            ]
        return self._sed

    @staticmethod
    def _round_sersic(n, sersic_prec):
        return float(int(n/sersic_prec + 0.5)) * sersic_prec

    @staticmethod
    def _buildParametric(record, sersic_prec, gsparams, chromatic, bandpass=None, sed=None):
        # Get fit parameters.  For 'sersicfit', the result is an array of 8 numbers for each
        # galaxy:
        #     SERSICFIT[0]: intensity of light profile at the half-light radius.
        #     SERSICFIT[1]: half-light radius measured along the major axis, in units of pixels
        #                   in the COSMOS lensing data reductions (0.03 arcsec).
        #     SERSICFIT[2]: Sersic n.
        #     SERSICFIT[3]: q, the ratio of minor axis to major axis length.
        #     SERSICFIT[4]: boxiness, currently fixed to 0, meaning isophotes are all
        #                   elliptical.
        #     SERSICFIT[5]: x0, the central x position in pixels.
        #     SERSICFIT[6]: y0, the central y position in pixels.
        #     SERSICFIT[7]: phi, the position angle in radians.  If phi=0, the major axis is
        #                   lined up with the x axis of the image.
        # For 'bulgefit', the result is an array of 16 parameters that comes from doing a
        # 2-component sersic fit.  The first 8 are the parameters for the disk, with n=1, and
        # the last 8 are for the bulge, with n=4.
        bparams = record['bulgefit']
        sparams = record['sersicfit']
        if 'hlr' not in record and self.use_sample is not None:  # pragma: no cover
            raise OSError("You still have the old COSMOS catalog.  Run the program "
                          "`galsim_download_cosmos -s %s` to upgrade."%(self.use_sample))

        use_bulgefit = record['use_bulgefit']
        if not use_bulgefit and not record['viable_sersic']:  # pragma: no cover
            # This shouldn't be possible I think...
            raise GalSimError("Cannot make parametric model for this galaxy!")

        if use_bulgefit:
            # Bulge parameters:
            # Minor-to-major axis ratio:
            bulge_q = bparams[11]
            # Position angle, now represented as a galsim.Angle:
            bulge_beta = bparams[15]*radians
            disk_q = bparams[3]
            disk_beta = bparams[7]*radians
            bulge_hlr = record['hlr'][1]
            bulge_flux = record['flux'][1]
            disk_hlr = record['hlr'][2]
            disk_flux = record['flux'][2]

            # Make sure the bulge-to-total flux ratio is not nonsense.
            bfrac = bulge_flux/(bulge_flux+disk_flux)
            if bfrac < 0 or bfrac > 1 or np.isnan(bfrac):  # pragma: no cover
                # This shouldn't be possible I think...
                raise GalSimError("Cannot make parametric model for this galaxy")

            # Then combine the two components of the galaxy.
            if chromatic:
                # We define the GSObjects with flux=1, then multiply by an SED defined to have
                # the appropriate (observed) magnitude at the redshift in the COSMOS passband.
                z = record['zphot']
                target_bulge_mag = record['mag_auto']-2.5*math.log10(bfrac)
                bulge_sed = sed[0].atRedshift(z).withMagnitude(
                    target_bulge_mag, bandpass)
                bulge = DeVaucouleurs(half_light_radius=bulge_hlr, gsparams=gsparams)
                bulge *= bulge_sed
                target_disk_mag = record['mag_auto']-2.5*math.log10((1.-bfrac))
                disk_sed = sed[1].atRedshift(z).withMagnitude(target_disk_mag, bandpass)
                disk = Exponential(half_light_radius=disk_hlr, gsparams=gsparams)
                disk *= disk_sed
            else:
                bulge = DeVaucouleurs(flux=bulge_flux, half_light_radius=bulge_hlr,
                                             gsparams=gsparams)
                disk = Exponential(flux=disk_flux, half_light_radius=disk_hlr,
                                          gsparams=gsparams)

            # Apply shears for intrinsic shape.
            if bulge_q < 1.:  # pragma: no branch
                bulge = bulge.shear(q=bulge_q, beta=bulge_beta)
            if disk_q < 1.:  # pragma: no branch
                disk = disk.shear(q=disk_q, beta=disk_beta)

            gal = bulge + disk
        else:
            # Do a similar manipulation to the stored quantities for the single Sersic profiles.
            gal_n = sparams[2]
            # Fudge this if it is at the edge of the allowed n values.  Since GalSim (as of #325 and
            # #449) allow Sersic n in the range 0.3<=n<=6, the only problem is that the fits
            # occasionally go as low as n=0.2.  The fits in this file only go to n=6, so there is no
            # issue with too-high values, but we also put a guard on that side in case other samples
            # are swapped in that go to higher value of sersic n.
            if gal_n < 0.3: gal_n = 0.3
            if gal_n > 6.0: gal_n = 6.0
            # GalSim is much more efficient if only a finite number of Sersic n values are used.
            # This (optionally given constructor args) rounds n to the nearest 0.05.
            if sersic_prec > 0.:
                gal_n = COSMOSCatalog._round_sersic(gal_n, sersic_prec)
            gal_q = sparams[3]
            gal_beta = sparams[7]*radians
            gal_hlr = record['hlr'][0]
            gal_flux = record['flux'][0]

            if chromatic:
                gal = Sersic(gal_n, flux=1., half_light_radius=gal_hlr, gsparams=gsparams)
                if gal_n < 1.5:
                    use_sed = sed[1] # disk
                elif gal_n >= 1.5 and gal_n < 3.0:
                    use_sed = sed[2] # intermediate
                else:
                    use_sed = sed[0] # bulge
                target_mag = record['mag_auto']
                z = record['zphot']
                gal *= use_sed.atRedshift(z).withMagnitude(target_mag, bandpass)
            else:
                gal = Sersic(gal_n, flux=gal_flux, half_light_radius=gal_hlr, gsparams=gsparams)

            # Apply shears for intrinsic shape.
            if gal_q < 1.:  # pragma: no branch
                gal = gal.shear(q=gal_q, beta=gal_beta)

        return gal

    def getRealParams(self, index):
        """Get the parameters needed to make a `RealGalaxy` for a given index."""
        # Used by COSMOSGalaxy to circumvent making the RealGalaxy here and potentially having
        # to pickle the result.  These raw materials should be smaller, so quicker to pickle.
        orig_index = self.orig_index[index]
        gal_image = self.real_cat.getGalImage(orig_index)
        psf_image = self.real_cat.getPSFImage(orig_index)
        noise_image, pixel_scale, var = self.real_cat.getNoiseProperties(orig_index)
        return (gal_image, psf_image, noise_image, pixel_scale, var)

    def getParametricRecord(self, index):
        """Get the parametric record for a given index"""
        # Used by _makeGalaxy to circumvent pickling the result.
        record = self.param_cat[self.orig_index[index]]
        # Convert to a dict, since on some systems, the numpy record doesn't seem to
        # pickle correctly.
        record_dict = { k:record[k] for k in record.dtype.names }
        return record_dict

    def getValue(self, key, index):
        """Get the value in the parametric catalog at the given key and index"""
        # Used by _makeGalaxy to circumvent pickling the result.
        record = self.param_cat[self.orig_index[index]]
        return record[key]

    def canMakeReal(self):
        """Is it permissible to call makeGalaxy with gal_type='real'?"""
        return self.use_real

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, GalaxySample) and
                 self.use_real == other.use_real and
                 self.real_cat == other.real_cat and
                 np.array_equal(self.param_cat, other.param_cat) and
                 np.array_equal(self.orig_index, other.orig_index)))

class COSMOSCatalog(GalaxySample):
    """
    A class representing a random subsample of galaxies from the COSMOS sample with F814W<25.2
    (default), or alternatively the entire sample with F814W<23.5.

    Depending on the keyword arguments, particularly ``use_real``, the catalog may be able to
    generate real galaxies, parametric galaxies, or both.  To use this with either type of
    galaxies, you need to get the COSMOS datasets in the format that GalSim recognizes; see

    https://github.com/GalSim-developers/GalSim/wiki/RealGalaxy-Data

    option (1) for more information.  Note that if you want to make real galaxies you need to
    download and store the full tarball with all galaxy images, whereas if you want to make
    parametric galaxies you only need the catalog real_galaxy_catalog_25.2_fits.fits (and the
    selection file real_galaxy_catalog_25.2_selection.fits if you want to place cuts on the
    postage stamp quality) and can delete the galaxy and PSF image files.

    Finally, we provide a program that will download the large COSMOS sample for you and
    put it in the $PREFIX/share/galsim directory of your installation path.  The program is::

            galsim_download_cosmos

    which gets installed in the $PREFIX/bin directory when you install GalSim.  If you use
    this program to download the COSMOS catalog, then you can use it with::

            cat = galsim.COSMOSCatalog()

    GalSim knows the location of the installation share directory, so it will automatically
    look for it there.

    In addition to the option of specifying catalog names, this class also accepts a keyword
    argument ``sample`` that can be used to switch between the samples with limiting magnitudes of
    23.5 and 25.2.

    After getting the catalogs, there is a method makeGalaxy() that can make a `GSObject`
    corresponding to any chosen galaxy in the catalog (whether real or parametric).  See
    `GalaxySample.makeGalaxy` for more information.  As an interesting application and example of
    the usage of these routines, consider the following code::

        >>> im_size = 64
        >>> pix_scale = 0.05
        >>> bp_file = os.path.join(galsim.meta_data.share_dir, 'wfc_F814W.dat.gz')
        >>> bandpass = galsim.Bandpass(bp_file, wave_type='ang').thin().withZeropoint(25.94)
        >>> cosmos_cat = galsim.COSMOSCatalog()
        >>> psf = galsim.OpticalPSF(diam=2.4, lam=1000.) # bigger than HST F814W PSF.
        >>> indices = np.arange(10)
        >>> real_gal_list = cosmos_cat.makeGalaxy(indices, gal_type='real',
        ...                                       noise_pad_size=im_size*pix_scale)
        >>> param_gal_list = cosmos_cat.makeGalaxy(indices, gal_type='parametric', chromatic=True)
        >>> for ind in indices:
        >>>     real_gal = galsim.Convolve(real_gal_list[ind], psf)
        >>>     param_gal = galsim.Convolve(param_gal_list[ind], psf)
        >>>     im_real = galsim.Image(im_size, im_size)
        >>>     im_param = galsim.Image(im_size, im_size)
        >>>     real_gal.drawImage(image=im_real, scale=pix_scale)
        >>>     param_gal.drawImage(bandpass, image=im_param, scale=pix_scale)
        >>>     im_real.write('im_real_'+str(ind)+'.fits')
        >>>     im_param.write('im_param_'+str(ind)+'.fits')

    This code snippet will draw images of the first 10 entries in the COSMOS catalog, at slightly
    lower resolution than in COSMOS, with a real image and its parametric representation for each of
    those objects.

    When using the 'real' rather than 'parametric' option, please read the documentation for the
    `RealGalaxy` class for additional caveats about the available drawing methods and
    the need to convolve with a suitable PSF.

    Parameters:
        file_name:          The file containing the catalog. [default: None, which will look for the
                            F814W<25.2 COSMOS catalog in $PREFIX/share/galsim.  It will raise an
                            exception if the catalog is not there telling you to run
                            galsim_download_cosmos.]
        sample:             A keyword argument that can be used to specify the sample to use, i.e.,
                            "23.5" or "25.2".  At most one of ``file_name`` and ``sample`` should be
                            specified.  [default: None, which results in the same default as
                            ``file_name=None``.]
        dir:                The directory with the catalog file and, if making realistic galaxies,
                            the image and noise files (or symlinks to them). [default: None, which
                            will look in $PREFIX/share/galsim.]
        preload:            Keyword that is only used for real galaxies, not parametric ones, to
                            choose whether to preload the header information.  If ``preload=True``,
                            the bulk of the I/O time is in the constructor.  If ``preload=False``,
                            there is approximately the same total I/O time (assuming you eventually
                            use most of the image files referenced in the catalog), but it is spread
                            over the calls to makeGalaxy().  [default: False]
        use_real:           Enable the use of realistic galaxies?  [default: True]
                            If this parameter is False, then ``makeGalaxy(gal_type='real')`` will
                            not be allowed, and there will be a (modest) decrease in RAM and time
                            spent on I/O when initializing the COSMOSCatalog. If the real
                            catalog is not available for some reason, it will still be possible to
                            make parametric images.
        exclusion_level:    Level of additional cuts to make on the galaxies based on the quality
                            of postage stamp definition and/or parametric fit quality [beyond the
                            minimal cuts imposed when making the catalog - see Mandelbaum et
                            al. (2012, MNRAS, 420, 1518) for details].  Options:

                            - "none": No cuts.
                            - "bad_stamp": Apply cuts to eliminate galaxies that have failures in
                              postage stamp definition.  These cuts may also eliminate a small
                              subset of the good postage stamps as well.
                            - "bad_fits": Apply cuts to eliminate galaxies that have failures in the
                              parametric fits.  These cuts may also eliminate a small
                              subset of the good parametric fits as well.
                            - "marginal": Apply the above cuts, plus ones that eliminate some more
                              marginal cases.

                            [default: "marginal"]
        min_hlr:            Exclude galaxies whose fitted half-light radius is smaller than this
                            value (in arcsec).  [default: 0, meaning no limit]
        max_hlr:            Exclude galaxies whose fitted half-light radius is larger than this
                            value (in arcsec).  [default: 0, meaning no limit]
        min_flux:           Exclude galaxies whose fitted flux is smaller than this value.
                            [default: 0, meaning no limit]
        max_flux:           Exclude galaxies whose fitted flux is larger than this value.
                            [default: 0, meaning no limit]
        exptime:            The exposure time (in seconds) to assume when creating galaxies.
                            .. note::

                                The processed COSMOS ACS/HST science images have units of
                                counts/second; i.e. they have an effective exposure time of 1
                                second in terms of their flux levels. The default value
                                corresponds to a 1 second exposure on HST, which will match
                                these processed images.

                            [default: 1]
        area:               The effective collecting area (in cm^2) to assume when creating
                            galaxies. [default: None, which means to use the original HST
                            collecting area = pi/4 * 240**2 * (1.-0.33**2)]

    After construction, the following attributes are available:

    Attributes:
        nobjects:       The number of objects in the catalog
    """
    _opt_params = { 'file_name' : str, 'sample' : str, 'dir' : str,
                    'preload' : bool, 'use_real' : bool,
                    'exclusion_level' : str, 'min_hlr' : float, 'max_hlr' : float,
                    'min_flux' : float, 'max_flux' : float
                  }

    hst_eff_area = math.pi * 120**2 * (1-0.33**2)

    def __init__(self, file_name=None, sample=None, dir=None, **kwargs):
        if sample is not None and file_name is not None:
            raise GalSimIncompatibleValuesError(
                "Cannot specify both the sample and file_name.",
                sample=sample, file_name=file_name)

        # Parse the file name
        file_name, _, use_sample = _parse_files_dirs(file_name, dir, sample)

        if use_sample == "23.5":
            cut_ratio = 0.2
            sn_limit = 20.0
            min_mask_dist = 11.
        else:
            cut_ratio = 0.8
            sn_limit = 12.0
            min_mask_dist = 11.

        super().__init__(file_name, _use_sample=use_sample,
                         orig_exptime=1., orig_area=self.hst_eff_area,
                         cut_ratio=cut_ratio, sn_limit=sn_limit, min_mask_dist=min_mask_dist,
                         **kwargs)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np

import galsim
from galsim_test_helpers import *


test_sigma = 7.                   # test value of Gaussian sigma for integral tests
test_rel_err = 1.e-7              # the relative accuracy at which to test
test_abs_err = 1.e-13             # the absolute accuracy at which to test
test_mock_inf = 2.e10             # number large enough to get interpreted as infinity by
                                  # integration routines
test_decimal = 7


@timer
def test_gaussian_finite_limits():
    """Test the integration of a 1D zero-mean Gaussian across intervals of [-1, 1], [0, 20]
    and [-50, -40].
    """
    # Define our test function
    def test_func(x): return np.exp(-.5 * x**2 / test_sigma**2)

    test_integral = galsim.integ.int1d(test_func, -1., 1., test_rel_err, test_abs_err)
    # test results easily calculated using Wolfram alpha
    true_result = 1.99321805307377285009
    np.testing.assert_almost_equal(
        test_integral, true_result, decimal=test_decimal, verbose=True,
        err_msg="Gaussian integral failed across interval [-1, 1].")

    test_integral = galsim.integ.int1d(test_func, 0., 20., test_rel_err, test_abs_err)
    true_result = 8.73569586966967345835
    np.testing.assert_almost_equal(
        test_integral, true_result, decimal=test_decimal, verbose=True,
        err_msg="Gaussian integral failed across interval [0, 20].")

    test_integral = galsim.integ.int1d(test_func, -50., -40., test_rel_err, test_abs_err)
    true_result = 9.66426031085587421984e-8
    np.testing.assert_almost_equal(
        test_integral, true_result, decimal=test_decimal, verbose=True,
        err_msg="Gaussian integral failed across interval [-50, -40].")


@timer
def test_gaussian_infinite_limits():
    """Test the integration of a 1D zero-mean Gaussian across intervals of [0, inf], [-inf, 5.4]
    and [-inf, inf].
    """
    # Define our test function
    def test_func(x): return np.exp(-.5 * x**2 / test_sigma**2)

    test_integral = galsim.integ.int1d(test_func, 0., test_mock_inf, test_rel_err, test_abs_err)
    # test results easily calculated using Wolfram alpha
    true_result = 8.77319896120850210849
    np.testing.assert_almost_equal(
        test_integral, true_result, decimal=test_decimal, verbose=True,
        err_msg="Gaussian integral failed across interval [0, inf].")

    test_integral = galsim.integ.int1d(test_func, -test_mock_inf, 5.4, test_rel_err, test_abs_err)
    true_result = 13.68221660030048620971
    np.testing.assert_almost_equal(
        test_integral, true_result, decimal=test_decimal, verbose=True,
        err_msg="Gaussian integral failed across interval [-inf, 5.4].")

    test_integral = galsim.integ.int1d(
        test_func, -test_mock_inf, test_mock_inf, test_rel_err, test_abs_err)
    true_result = 17.54639792241700421699
    np.testing.assert_almost_equal(
        test_integral, true_result, decimal=test_decimal, verbose=True,
        err_msg="Gaussian integral failed across interval [-inf, inf].")


@timer
def test_sinxsqexpabsx_finite_limits():
    """Test the integration of a slightly tricky oscillating sin(x^2) * exp(-|x|) function across
    finite intervals [-1, 1], [0, 20], [-15, 14].
    """
    # Define our test function
    def test_func(x): return np.sin(x**2) * np.exp(-np.abs(x))

    test_integral = galsim.integ.int1d(test_func, -1., 1., test_rel_err, test_abs_err)
    # test results easily calculated using Wolfram alpha
    true_result = 0.30182513444548879567
    np.testing.assert_almost_equal(
        test_integral, true_result, decimal=test_decimal, verbose=True,
        err_msg="Sin(x^2) * exp(-|x|) integral failed across interval [-1, 1].")

    test_integral = galsim.integ.int1d(test_func, 0., 20., test_rel_err, test_abs_err)
    true_result = 0.27051358019041255485
    np.testing.assert_almost_equal(
        test_integral, true_result, decimal=test_decimal, verbose=True,
        err_msg="Sin(x^2) * exp(-|x|) integral failed across interval [0, 20].")

    test_integral = galsim.integ.int1d(test_func, -15., -14., test_rel_err, test_abs_err)
    true_result = 7.81648378350593176887e-9
    np.testing.assert_almost_equal(
        (test_integral - true_result) / true_result, 0., decimal=test_decimal, verbose=True,
        err_msg="Sin(x^2) * exp(-|x|) integral failed across interval [-15, -14].")


@timer
def test_sinxsqexpabsx_infinite_limits():
    """Test the integration of a slightly tricky oscillating sin(x^2) * exp(-|x|) function across
    infinite intervals [0, inf], [-inf, 5.4], [-inf, inf].
    """
    # Define our test function
    def test_func(x): return np.sin(x**2) * np.exp(-np.abs(x))

    test_integral = galsim.integ.int1d(test_func, 0., test_mock_inf, test_rel_err, test_abs_err)
    # test results easily calculated using Wolfram alpha
    true_result = 0.27051358016221414426
    np.testing.assert_almost_equal(
        test_integral, true_result, decimal=test_decimal, verbose=True,
        err_msg="Sin(x^2) * exp(-|x|) integral failed across interval [0, inf].")

    test_integral = galsim.integ.int1d(test_func, -test_mock_inf, 5.4, test_rel_err, test_abs_err)
    true_result = 0.5413229824941895221
    np.testing.assert_almost_equal(
        test_integral, true_result, decimal=test_decimal, verbose=True,
        err_msg="Sin(x^2) * exp(-|x|) integral failed across interval [-inf, 5.4].")

    test_integral = galsim.integ.int1d(
        test_func, -test_mock_inf, test_mock_inf, test_rel_err, test_abs_err)
    true_result = 0.54102716032442828852
    np.testing.assert_almost_equal(
        test_integral, true_result, decimal=test_decimal, verbose=True,
        err_msg="Sin(x^2) * exp(-|x|) integral failed across interval [-inf, inf].")


@timer
def test_invroot_finite_limits():
    """Test the integration of |x|^(-1/2) across intervals [0,1], [0,300] (integrable pole at x=0).
    """
    # Define our test function
    def test_func(x): return 1. / np.sqrt(np.abs(x))
    test_integral = galsim.integ.int1d(test_func, 0, 1., test_rel_err, test_abs_err)
    # test results easily calculated using Wolfram alpha
    true_result = 2.
    np.testing.assert_almost_equal(
        test_integral, true_result, decimal=test_decimal, verbose=True,
        err_msg="|x|^(-1/2) integral failed across interval [0, 1].")

    test_integral = galsim.integ.int1d(test_func, 0., 300., test_rel_err, test_abs_err)
    true_result = 34.64101615137754587055
    np.testing.assert_almost_equal(
        test_integral, true_result, decimal=test_decimal, verbose=True,
        err_msg="|x|^(-1/2) integral failed across interval [0, 300].")


@timer
def test_invroot_infinite_limits():
    """Test the integration of |x|^(-2) across intervals [1,2], [1,inf].
    Also check that [0,1] raises an exception.
    """
    # Define our test function
    def test_func(x): return x**-2
    test_integral = galsim.integ.int1d(test_func, 1., 2., test_rel_err, test_abs_err)
    true_result = 0.5
    np.testing.assert_almost_equal(
        test_integral, true_result, decimal=test_decimal, verbose=True,
        err_msg="x^(-2) integral failed across interval [1, 2].")

    test_integral = galsim.integ.int1d(test_func, 1., test_mock_inf, test_rel_err, test_abs_err)
    true_result = 1.0
    np.testing.assert_almost_equal(
        test_integral, true_result, decimal=test_decimal, verbose=True,
        err_msg="x^(-2) integral failed across interval [1, inf].")

    with assert_raises(galsim.GalSimError):
        galsim.integ.int1d(test_func, 0., 1., test_rel_err, test_abs_err)


@timer
def test_midpoint_basic():
    """Test the basic functionality of the midptRule() function.
    """
    # This shouldn't be super accurate, but just make sure it's not really broken.
    x = 0.01*np.arange(1000)
    func = lambda x:x**2
    result = galsim.integ.midptRule(func, x)
    expected_val = 1.e3/3.
    np.testing.assert_almost_equal(
        result/expected_val, 1.0, decimal=2, verbose=True,
        err_msg='Simple test of midptRule() method failed for f(x)=x^2 from 0 to 10')

    # Error with 0 or 1 ascissae
    with assert_raises(galsim.GalSimValueError):
        galsim.integ.midptRule(func, [0])
    with assert_raises(galsim.GalSimValueError):
        galsim.integ.midptRule(func, [])

@timer
def test_trapz_basic():
    """Test the basic functionality of the trapzRule() function.
    """
    # This shouldn't be super accurate, but just make sure it's not really broken.
    x = np.linspace(0, 1, 100000)
    func = lambda x: x**2
    result = galsim.integ.trapzRule(func, x)
    expected_val = 1./3.
    np.testing.assert_almost_equal(
        result/expected_val, 1.0, decimal=6, verbose=True,
        err_msg='Test of trapzRule() with points failed for f(x)=x^2 from 0 to 1')

    # quadRule with no weight is equivalent.
    result2 = galsim.integ.quadRule(func, x)
    np.testing.assert_almost_equal(result2, result)

    # Error with 0 or 1 ascissae
    with assert_raises(galsim.GalSimValueError):
        galsim.integ.trapzRule(func, [0])
    with assert_raises(galsim.GalSimValueError):
        galsim.integ.trapzRule(func, [])

@timer
def test_quad_basic():
    """Test the basic functionality of the quadRule() function.
    """
    x = np.linspace(0, 1, 1000)
    # With func and weight both linear, quad should be exact.
    func = lambda x: x
    weight = lambda x: 1+x
    result1 = galsim.integ.midptRule(func, x, weight)
    result2 = galsim.integ.trapzRule(func, x, weight)
    result3 = galsim.integ.quadRule(func, x, weight)
    expected_val = 5./6.
    print('int(x^2 (1+x), 0..1) = ')
    print('midpt: ',result1)
    print('trapz: ',result2)
    print('quad:  ',result3)
    print('true:  ',expected_val)
    np.testing.assert_allclose(result1, expected_val, rtol=2.e-3)
    np.testing.assert_allclose(result2, expected_val, rtol=3.e-7)
    np.testing.assert_allclose(result3, expected_val, rtol=1.e-14)

    # Check only 2 abscissae
    x = np.linspace(0, 1, 2)
    func = lambda x: x
    result1 = galsim.integ.midptRule(func, x, weight)
    result2 = galsim.integ.trapzRule(func, x, weight)
    result3 = galsim.integ.quadRule(func, x, weight)
    np.testing.assert_allclose(result1, expected_val, rtol=2)
    np.testing.assert_allclose(result2, expected_val, rtol=0.3)
    np.testing.assert_allclose(result3, expected_val, rtol=1.e-10)

    # Error with 0 or 1 abscissae
    with assert_raises(galsim.GalSimError):
        galsim.integ.quadRule(func, [0], weight)
    with assert_raises(galsim.GalSimError):
        galsim.integ.quadRule(func, [], weight)
    with assert_raises(galsim.GalSimError):
        galsim.integ.quadRule(func, [0])
    with assert_raises(galsim.GalSimError):
        galsim.integ.quadRule(func, [])

@timer
def test_hankel():
    """Test the galsim.integ.hankel function
    """
    # Most of the use of this function is in C++, but we provide hooks to use it in python
    # too in case that's useful for people.
    # (We don't currently use this from python in GalSim proper.)

    f1 = lambda r: np.exp(-r)
    # The Hankel transform of exp(-r) is (1+k^2)**(-3/2)
    for k in [1, 1.e-2, 0.234, 23.9, 1.e-8, 0]:
        result = galsim.integ.hankel(f1, k)
        expected_val = (1+k**2)**-1.5
        np.testing.assert_allclose(result, expected_val)

    r0 = 1.7
    f2 = lambda r: 1.-(r/r0)**2
    # The truncated Hankel transform of (1-(r/r0)^2) up to r0 is 2 J_2(r0 k)/k^2
    for k in [1, 1.e-2, 0.234, 23.9, 1.e-8, 0]:
        result = galsim.integ.hankel(f2, k, rmax=r0, rel_err=1.e-8)
        if k == 0:
            expected_val = r0**2/4  # The limit as k->0.
        else:
            expected_val = 2*galsim.bessel.jn(2,r0*k)/k**2
        np.testing.assert_allclose(result, expected_val, rtol=1.e-6)

    # The generalized Hankel transform of exp(-r) is:
    # (1 + nu (1+k^2)^0.5) k^nu / ( (1+k^2)^1.5 (1 + (1+k^2)^0.5)^nu
    for nu in [0, 1, 7, 0.5, 0.003, 12.23]:
        for k in [1, 1.e-2, 0.234, 23.9, 1.e-5, 1.e-8, 0]:
            result = galsim.integ.hankel(f1, k, nu=nu)
            expected_val = (1+k**2)**-1.5 * (1+nu*(1+k**2)**0.5) * k**nu / (1+(1+k**2)**0.5)**nu
            print(nu, k, result, expected_val)
            np.testing.assert_allclose(result, expected_val, rtol=1.e-6, atol=1.e-12)

    # Test doing multiple k values at once.
    k = np.array([1, 1.e-2, 0.234, 23.9, 1.e-5, 1.e-8, 0])
    for nu in [0, 1, 7, 0.5, 0.003, 12.23]:
        result = galsim.integ.hankel(f1, k, nu=nu)
        expected_val = (1+k**2)**-1.5 * (1+nu*(1+k**2)**0.5) * k**nu / (1+(1+k**2)**0.5)**nu
        np.testing.assert_allclose(result, expected_val, rtol=1.e-6, atol=1.e-12)

    with assert_raises(galsim.GalSimValueError):
        galsim.integ.hankel(f1, k=-0.3)
    with assert_raises(galsim.GalSimValueError):
        galsim.integ.hankel(f1, k=0.3, nu=-1)
    with assert_raises(galsim.GalSimValueError):
        galsim.integ.hankel(f1, k=0.3, nu=-0.5)


@timer
def test_gq_annulus():
    """Test the galsim.integ.gq_annulus function
    """

    # We can use the normalization of annular Zernikes to test.
    # From the Zernike docs:
    # \int_\mathrm{annulus} Z_i Z_j dA =
    # \pi \left(R_\mathrm{outer}^2 - R_\mathrm{inner}^2\right) \delta_{i, j}
    # I.e., the integral is the annulus area times Kronecker delta.

    rng = galsim.BaseDeviate(1111).as_numpy_generator()
    for j1 in range(1, 28):   # up to 6th order xy polynomials
        r1 = rng.uniform(0.3, 0.6)
        r2 = rng.uniform(0.9, 1.1)
        area = np.pi*(r2**2 - r1**2)
        Z1 = galsim.zernike.Zernike([0]*j1+[1], R_inner=r1, R_outer=r2)
        for j2 in range(1, 28):
            Z2 = galsim.zernike.Zernike([0]*j2+[1], R_inner=r1, R_outer=r2)
            # Product of 2 6th order polynomials is 12th order, so we need 6 rings and 13 spokes.
            np.testing.assert_allclose(
                galsim.integ.gq_annulus(Z1*Z2, r2, r1, n_rings=6, n_spokes=13),
                area if j1 == j2 else 0,
                rtol=1.e-11, atol=1.e-11
            )


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

import numpy as np

import galsim
from galsim_test_helpers import *


@timer
def test_shear_position():
    # zero position
    shear = galsim.Shear(g1=-0.03, g2=0.7)
    pos = galsim.PositionD(x=0, y=0)
    sheared_pos = pos.shear(shear)
    np.testing.assert_almost_equal(
        sheared_pos.x, 0,
        err_msg="x coord is wrong for shear with zero pos"
    )
    np.testing.assert_almost_equal(
        sheared_pos.y, 0,
        err_msg="y coord is wrong for shear with zero pos"
    )

    # zero shear
    shear = galsim.Shear(g1=0, g2=0)
    pos = galsim.PositionD(x=0.1, y=-0.5)
    sheared_pos = pos.shear(shear)
    np.testing.assert_almost_equal(
        sheared_pos.x, pos.x,
        err_msg="x coord is wrong for zero shear with pos"
    )
    np.testing.assert_almost_equal(
        sheared_pos.y, pos.y,
        err_msg="y coord is wrong for zero shear with pos"
    )

    # full thing
    shear = galsim.Shear(g1=-0.1, g2=0.1)
    pos = galsim.PositionD(x=0.1, y=-0.5)
    sheared_pos = pos.shear(shear)
    mat = shear.getMatrix()
    np.testing.assert_almost_equal(
        sheared_pos.x,
        pos.x * mat[0, 0] + pos.y * mat[0, 1],
        err_msg="x coord is wrong for shear with pos"
    )
    np.testing.assert_almost_equal(
        sheared_pos.y,
        pos.x * mat[1, 0] + pos.y * mat[1, 1],
        err_msg="y coord is wrong for shear with pos"
    )


@timer
def test_shear_position_image_integration_pixelwcs():
    wcs = galsim.PixelScale(0.3)
    obj1 = galsim.Gaussian(sigma=3)
    obj2 = galsim.Gaussian(sigma=2)
    pos2 = galsim.PositionD(3, 5)
    sum = obj1 + obj2.shift(pos2)
    shear = galsim.Shear(g1=0.1, g2=0.18)
    im1 = galsim.Image(50, 50, wcs=wcs)
    sum.shear(shear).drawImage(im1, center=im1.center)

    # Equivalent to shear each object separately and drawing at the sheared position.
    im2 = galsim.Image(50, 50, wcs=wcs)
    obj1.shear(shear).drawImage(im2, center=im2.center)
    obj2.shear(shear).drawImage(
        im2,
        add_to_image=True,
        center=im2.center + wcs.toImage(pos2.shear(shear)),
    )

    print("err:", np.max(np.abs(im1.array - im2.array)))
    assert np.allclose(im1.array, im2.array, rtol=0, atol=5e-8)


@timer
def test_shear_position_image_integration_offsetwcs():
    wcs = galsim.OffsetShearWCS(
        0.3,
        galsim.Shear(g1=-0.2, g2=0.4),
        origin=galsim.PositionD(x=-1.1, y=2.7),
        world_origin=galsim.PositionD(x=0.6, y=-2.5),
    )

    obj1 = galsim.Gaussian(sigma=3)
    obj2 = galsim.Gaussian(sigma=2)
    pos2 = galsim.PositionD(3, 5)
    pos1 = galsim.PositionD(0, 0)
    sum_obj = obj1 + obj2.shift(pos2)
    shear = galsim.Shear(g1=0.1, g2=0.18)
    im1 = galsim.Image(50, 50, wcs=wcs)
    sum_obj.shear(shear).drawImage(im1, center=wcs.toImage(pos1))

    # Equivalent to shear each object separately and drawing at the sheared position.
    im2 = galsim.Image(50, 50, wcs=wcs)
    obj1.shear(shear).drawImage(im2, center=wcs.toImage(pos1))
    obj2.shear(shear).drawImage(
        im2,
        add_to_image=True,
        center=wcs.toImage(pos2.shear(shear)),
    )

    print("err:", np.max(np.abs(im1.array - im2.array)))
    assert np.allclose(im1.array, im2.array, rtol=0, atol=2e-7), (
        np.max(np.abs(im1.array - im2.array))
    )


if __name__ == "__main__":
    runtests(__file__)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'Position', 'PositionI', 'PositionD', '_PositionI', '_PositionD', ]

from . import _galsim

class Position:
    """A class for representing 2D positions on the plane.

    Position is a base class for two slightly different kinds of positions:
    PositionD describes positions with floating point values in ``x`` and ``y``.
    PositionI described positions with integer values in ``x`` and ``y``.

    In the C++ layer, these are templates, but of course no such thing exists in Python,
    so the trailing D or I indicate the type.

    Initialization:

    For the float-valued position class, example initializations include::

        >>> pos = galsim.PositionD(x=0.5, y=-0.5)
        >>> pos = galsim.PositionD(0.5, -0.5)
        >>> pos = galsim.PositionD( (0.5, -0.5) )

    And for the integer-valued position class, example initializations include::

        >>> pos = galsim.PositionI(x=45, y=13)
        >>> pos = galsim.PositionI(45, 13)
        >>> pos = galsim.PositionD( (45, 15) )

    Attributes:
        x:      The x component of the position
        y:      The y component of the position

    Arithmetic:

    Most arithmetic that makes sense for a position is allowed::

        >>> pos1 + pos2
        >>> pos1 - pos2
        >>> pos * x
        >>> pos / x
        >>> -pos
        >>> pos1 += pos2
        >>> pos1 -= pos2
        >>> pos *= x
        >>> pos -= x

    Note though that the types generally need to match.  For example, you cannot multiply
    a PositionI by a float add a PositionD to a PositionI in place.

    Position instances can be sheared by a `galsim.Shear`::

        >>> pos = galsim.PositionD(x=0.5, y=-0.5)
        >>> shear = galsim.Shear(g1=0.1, g2=-0.1)
        >>> sheared_pos = pos.shear(shear)

    Note that this operation will always return a PositionD even if
    an integer position is being sheared.
    """
    def __init__(self):
        raise NotImplementedError("Cannot instantiate the base class.  "
                                  "Use either PositionD or PositionI.")

    def _parse_args(self, *args, **kwargs):
        if len(kwargs) == 0:
            if len(args) == 2:
                self.x, self.y = args
            elif len(args) == 0:
                self.x = self.y = 0
            elif len(args) == 1:
                if isinstance(args[0], (Position, _galsim.PositionD, _galsim.PositionI)):
                    self.x = args[0].x
                    self.y = args[0].y
                else:
                    try:
                        self.x, self.y = args[0]
                    except (TypeError, ValueError):
                        raise TypeError("Single argument to %s must be either a Position "
                                        "or a tuple."%self.__class__) from None
            else:
                raise TypeError("%s takes at most 2 arguments (%d given)"%(
                        self.__class__, len(args)))
        elif len(args) != 0:
            raise TypeError("%s takes x and y as either named or unnamed arguments (given %s, %s)"%(
                    self.__class__, args, kwargs))
        else:
            try:
                self.x = kwargs.pop('x')
                self.y = kwargs.pop('y')
            except KeyError:
                raise TypeError(
                    "Keyword arguments x,y are required for %s"%self.__class__) from None
            if kwargs:
                raise TypeError("Got unexpected keyword arguments %s"%kwargs.keys())

    def __mul__(self, other):
        self._check_scalar(other, 'multiply')
        return self.__class__(self.x * other, self.y * other)

    def __rmul__(self, other):
        return self.__mul__(other)

    def __div__(self, other):
        self._check_scalar(other, 'divide')
        return self.__class__(self.x / other, self.y / other)

    __truediv__ = __div__

    def __neg__(self):
        return self.__class__(-self.x, -self.y)

    def __add__(self, other):
        if isinstance(other, Bounds):
            return other + self
        elif not isinstance(other,Position):
            raise TypeError("Can only add a Position to a %s"%self.__class__.__name__)
        elif isinstance(other, self.__class__):
            return self.__class__(self.x + other.x, self.y + other.y)
        else:
            return PositionD(self.x + other.x, self.y + other.y)

    def __sub__(self, other):
        if not isinstance(other,Position):
            raise TypeError("Can only subtract a Position from a %s"%self.__class__.__name__)
        elif isinstance(other, self.__class__):
            return self.__class__(self.x - other.x, self.y - other.y)
        else:
            return _PositionD(self.x - other.x, self.y - other.y)

    def __repr__(self):
        return "galsim.%s(x=%r, y=%r)"%(self.__class__.__name__, self.x, self.y)

    def __str__(self):
        return "galsim.%s(%s,%s)"%(self.__class__.__name__, self.x, self.y)

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, self.__class__) and self.x == other.x and self.y == other.y))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.__class__.__name__, self.x, self.y))

    def shear(self, shear):
        """Shear the position.

        See the doc string of `galsim.Shear.getMatrix` for more details.

        Parameters:
            shear:    a `galsim.Shear` instance

        Returns:
            a `galsim.PositionD` instance.
        """
        shear_mat = shear.getMatrix()
        return PositionD(
            self.x * shear_mat[0, 0] + self.y * shear_mat[0, 1],
            self.x * shear_mat[1, 0] + self.y * shear_mat[1, 1],
        )


class PositionD(Position):
    """A Position that takes floating point values.

    See the Position doc string for more details.
    """
    def __init__(self, *args, **kwargs):
        self._parse_args(*args, **kwargs)
        self.x = float(self.x)
        self.y = float(self.y)

    @property
    def _p(self):
        return _galsim.PositionD(self.x, self.y)

    def round(self):
        """Return the rounded-off PositionI version of this position.
        """
        return _PositionI(int(round(self.x)), int(round(self.y)))

    def _check_scalar(self, other, op):
        try:
            if other == float(other): return
        except (TypeError, ValueError):
            pass
        raise TypeError("Can only %s a PositionD by float values"%op)


class PositionI(Position):
    """A Position that takes only integer values.

    Typically used for coordinate positions on an image.

    See the Position doc string for more details.
    """
    def __init__(self, *args, **kwargs):
        self._parse_args(*args, **kwargs)
        if self.x != int(self.x) or self.y != int(self.y):
            raise TypeError("PositionI must be initialized with integer values")
        self.x = int(self.x)
        self.y = int(self.y)

    @property
    def _p(self):
        return _galsim.PositionI(self.x, self.y)

    def round(self):
        # Just for consistency between PositionD and PositionI
        return self

    def _check_scalar(self, other, op):
        try:
            if other == int(other): return
        except (TypeError, ValueError):
            pass
        raise TypeError("Can only %s a PositionI by integer values"%op)

def _PositionD(x, y):
    """Equivalent to `PositionD` constructor, but skips some sanity checks and argument parsing.
    This requires that x,y are floats.
    """
    ret = PositionD.__new__(PositionD)
    ret.x = float(x)
    ret.y = float(y)
    return ret


def _PositionI(x, y):
    """Equivalent to `PositionI` constructor, but skips some sanity checks and argument parsing.
    This requires that x,y are ints.
    """
    ret = PositionI.__new__(PositionI)
    ret.x = int(x)
    ret.y = int(y)
    return ret

def parse_pos_args(args, kwargs, name1, name2, integer=False, others=[]):
    """Parse the args and kwargs of a function call to be some kind of position.

    We allow four options:

        f(x,y)
        f(galsim.PositionD(x,y)) or f(galsim.PositionI(x,y))
        f( (x,y) )  (or any indexable thing)
        f(name1=x, name2=y)

    If the inputs must be integers, set ``integer=True``.
    If there are other args/kwargs to parse after these, then their names should be
    be given as the parameter ``others``, which are passed back in a tuple after the position.

    Parameters:
        args:       The args of the original function.
        kwargs:     The kwargs of the original function.
        name1:      The allowed kwarg for the first coordinate.
        name2:      The allowed kwarg for the second coordinate.
        integer:    Whether to return a `PositionI` rather than a `PositionD`. [default: False]
        others:     If given, other values to also parse and return from the kwargs. [default: []]

    Returns:
        a `Position` instance, possibly also with other values if ``others`` is given.
    """
    def canindex(arg):
        try: arg[0], arg[1]
        except (TypeError, IndexError): return False
        else: return True

    other_vals = []
    if len(args) == 0:
        # Then name1,name2 need to be kwargs
        try:
            x = kwargs.pop(name1)
            y = kwargs.pop(name2)
        except KeyError:
            raise TypeError(
                'Expecting kwargs %s, %s.  Got %s'%(name1, name2, kwargs.keys())) from None
    elif ( ( isinstance(args[0], PositionI) or
             (not integer and isinstance(args[0], PositionD)) ) and
           len(args) <= 1+len(others) ):
        x = args[0].x
        y = args[0].y
        for arg in args[1:]:
            other_vals.append(arg)
            others.pop(0)
    elif canindex(args[0]) and len(args) <= 1+len(others):
        x = args[0][0]
        y = args[0][1]
        for arg in args[1:]:
            other_vals.append(arg)
            others.pop(0)
    elif len(args) == 1:
        if integer:
            raise TypeError("Cannot parse argument %s as a PositionI"%(args[0]))
        else:
            raise TypeError("Cannot parse argument %s as a PositionD"%(args[0]))
    elif len(args) <= 2 + len(others):
        x = args[0]
        y = args[1]
        for arg in args[2:]:
            other_vals.append(arg)
            others.pop(0)
    else:
        raise TypeError("Too many arguments supplied")
    # Read any remaining other kwargs
    if others:
        for name in others:
            val = kwargs.pop(name)
            other_vals.append(val)
    if kwargs:
        raise TypeError("Received unexpected keyword arguments: %s",kwargs)

    if integer:
        pos = _PositionI(int(x),int(y))
    else:
        pos = _PositionD(float(x),float(y))
    if other_vals:
        return (pos,) + tuple(other_vals)
    else:
        return pos


# Put this at the bottom to avoid circular import error
from .bounds import Bounds



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'Bounds', 'BoundsI', 'BoundsD', '_BoundsI', '_BoundsD', ]

import math

from . import _galsim
from .position import *
from .errors import GalSimUndefinedBoundsError

class Bounds:
    """A class for representing image bounds as 2D rectangles.

    Bounds is a base class for two slightly different kinds of bounds:
    `BoundsD` describes bounds with floating point values in x and y.
    `BoundsI` described bounds with integer values in x and y.

    The bounds are stored as four numbers in each instance, (xmin, xmax, ymin, ymax), with an
    additional boolean switch to say whether or not the Bounds rectangle has been defined.  The
    rectangle is undefined if the min value > the max value in either direction.

    *Initialization*:

    A `BoundsI` or `BoundsD` instance can be initialized in a variety of ways.  The most direct is
    via four scalars::

        >>> bounds = galsim.BoundsD(xmin, xmax, ymin, ymax)
        >>> bounds = galsim.BoundsI(imin, imax, jmin, jmax)

    In the `BoundsI` example above, ``imin``, ``imax``, ``jmin`` and ``jmax`` must all be integers
    to avoid a TypeError exception.

    Another way to initialize a `Bounds` instance is using two `Position` instances, the first
    for ``(xmin,ymin)`` and the second for ``(xmax,ymax)``::

        >>> bounds = galsim.BoundsD(galsim.PositionD(xmin, ymin), galsim.PositionD(xmax, ymax))
        >>> bounds = galsim.BoundsI(galsim.PositionI(imin, jmin), galsim.PositionI(imax, jmax))

    In both the examples above, the I/D type of `PositionI`/`PositionD` must match that of
    `BoundsI`/`BoundsD`.

    Finally, there are a two ways to lazily initialize a bounds instance with ``xmin = xmax``,
    ``ymin = ymax``, which will have an undefined rectangle and the instance method isDefined()
    will return False.  The first sets ``xmin = xmax = ymin = ymax = 0``::

        >>> bounds = galsim.BoundsD()
        >>> bounds = galsim.BoundsI()

    The second method sets both upper and lower rectangle bounds to be equal to some position::

        >>> bounds = galsim.BoundsD(galsim.PositionD(xmin, ymin))
        >>> bounds = galsim.BoundsI(galsim.PositionI(imin, jmin))

    Once again, the I/D type of `PositionI`/`PositionD` must match that of `BoundsI`/`BoundsD`.

    For the latter two initializations, you would typically then add to the bounds with::

        >>> bounds += pos1
        >>> bounds += pos2
        >>> [etc.]

    Then the bounds will end up as the bounding box of all the positions that were added to it.

    You can also find the intersection of two bounds with the & operator::

        >>> overlap = bounds1 & bounds2

    This is useful for adding one image to another when part of the first image might fall off
    the edge of the other image::

        >>> overlap = stamp.bounds & image.bounds
        >>> image[overlap] += stamp[overlap]

    """
    def __init__(self):
        raise NotImplementedError("Cannot instantiate the base class. "
                                  "Use either BoundsD or BoundsI.")

    def _parse_args(self, *args, **kwargs):
        if len(kwargs) == 0:
            if len(args) == 4:
                self._isdefined = True
                self.xmin, self.xmax, self.ymin, self.ymax = args
            elif len(args) == 0:
                self._isdefined = False
                self.xmin = self.xmax = self.ymin = self.ymax = 0
            elif len(args) == 1:
                if isinstance(args[0], (Bounds, _galsim.BoundsD, _galsim.BoundsI)):
                    self._isdefined = True
                    self.xmin = args[0].xmin
                    self.xmax = args[0].xmax
                    self.ymin = args[0].ymin
                    self.ymax = args[0].ymax
                elif isinstance(args[0], (Position, _galsim.PositionD, _galsim.PositionI)):
                    self._isdefined = True
                    self.xmin = self.xmax = args[0].x
                    self.ymin = self.ymax = args[0].y
                else:
                    raise TypeError("Single argument to %s must be either a Bounds or a Position"%(
                                    self.__class__.__name__))
                self._isdefined = True
            elif len(args) == 2:
                if (isinstance(args[0], (Position, _galsim.PositionD, _galsim.PositionI)) and
                    isinstance(args[1], (Position, _galsim.PositionD, _galsim.PositionI))):
                    self._isdefined = True
                    self.xmin = min(args[0].x, args[1].x)
                    self.xmax = max(args[0].x, args[1].x)
                    self.ymin = min(args[0].y, args[1].y)
                    self.ymax = max(args[0].y, args[1].y)
                else:
                    raise TypeError("Two arguments to %s must be Positions"%(
                                    self.__class__.__name__))
            else:
                raise TypeError("%s takes either 1, 2, or 4 arguments (%d given)"%(
                                self.__class__.__name__,len(args)))
        elif len(args) != 0:
            raise TypeError("Cannot provide both keyword and non-keyword arguments to %s"%(
                            self.__class__.__name__))
        else:
            try:
                self._isdefined = True
                self.xmin = kwargs.pop('xmin')
                self.xmax = kwargs.pop('xmax')
                self.ymin = kwargs.pop('ymin')
                self.ymax = kwargs.pop('ymax')
            except KeyError:
                raise TypeError("Keyword arguments, xmin, xmax, ymin, ymax are required for %s"%(
                                self.__class__.__name__)) from None
            if kwargs:
                raise TypeError("Got unexpected keyword arguments %s"%kwargs.keys())

        if not (float(self.xmin) <= float(self.xmax) and float(self.ymin) <= float(self.ymax)):
            self._isdefined = False

    def area(self):
        """Return the area of the enclosed region.

        The area is a bit different for integer-type `BoundsI` and float-type `BoundsD` instances.
        For floating point types, it is simply ``(xmax-xmin)*(ymax-ymin)``.  However, for integer
        types, we add 1 to each size to correctly count the number of pixels being described by the
        bounding box.
        """
        return self._area()

    def withBorder(self, dx, dy=None):
        """Return a new `Bounds` object that expands the current bounds by the specified width.

        If two arguments are given, then these are separate dx and dy borders.
        """
        self._check_scalar(dx, "dx")
        if dy is None:
            dy = dx
        else:
            self._check_scalar(dy, "dy")
        return self.__class__(self.xmin-dx, self.xmax+dx, self.ymin-dy, self.ymax+dy)

    @property
    def origin(self):
        "The lower left position of the `Bounds`."
        return self._pos_class(self.xmin, self.ymin)

    @property
    def center(self):
        """The central position of the `Bounds`.

        For a `BoundsI`, this will return an integer `PositionI`, which will be above and/or to
        the right of the true center if the x or y ranges have an even number of pixels.

        For a `BoundsD`, this is equivalent to true_center.
        """
        if not self.isDefined():
            raise GalSimUndefinedBoundsError("center is invalid for an undefined Bounds")
        return self._center

    @property
    def true_center(self):
        """The central position of the `Bounds` as a `PositionD`.

        This is always (xmax + xmin)/2., (ymax + ymin)/2., even for integer `BoundsI`, where
        this may not necessarily be an integer `PositionI`.
        """
        if not self.isDefined():
            raise GalSimUndefinedBoundsError("true_center is invalid for an undefined Bounds")
        return _PositionD((self.xmax + self.xmin)/2., (self.ymax + self.ymin)/2.)

    def includes(self, *args):
        """Test whether a supplied ``(x,y)`` pair, `Position`, or `Bounds` lie within a defined
        `Bounds` rectangle of this instance.

        Examples::

            >>> bounds = galsim.BoundsD(0., 100., 0., 100.)
            >>> bounds.includes(50., 50.)
            True
            >>> bounds.includes(galsim.PositionD(50., 50.))
            True
            >>> bounds.includes(galsim.BoundsD(-50., -50., 150., 150.))
            False

        The type of the `PositionI`/`PositionD` and `BoundsI`/`BoundsD` instances (i.e. integer or
        float type) should match that of the bounds instance.
        """
        if len(args) == 1:
            if isinstance(args[0], Bounds):
                b = args[0]
                return (self.isDefined() and b.isDefined() and
                        self.xmin <= b.xmin and
                        self.xmax >= b.xmax and
                        self.ymin <= b.ymin and
                        self.ymax >= b.ymax)
            elif isinstance(args[0], Position):
                p = args[0]
                return (self.isDefined() and
                        self.xmin <= p.x <= self.xmax and
                        self.ymin <= p.y <= self.ymax)
            else:
                raise TypeError("Invalid argument %s"%args[0])
        elif len(args) == 2:
            x, y = args
            return (self.isDefined() and
                    self.xmin <= float(x) <= self.xmax and
                    self.ymin <= float(y) <= self.ymax)
        elif len(args) == 0:
            raise TypeError("include takes at least 1 argument (0 given)")
        else:
            raise TypeError("include takes at most 2 arguments (%d given)"%len(args))

    def expand(self, factor_x, factor_y=None):
        """Grow the `Bounds` by the supplied factor about the center.

        If two arguments are given, then these are separate x and y factors to
        expand by.
        """
        if factor_y is None:
            factor_y = factor_x
        dx = (self.xmax - self.xmin) * 0.5 * (factor_x-1.)
        dy = (self.ymax - self.ymin) * 0.5 * (factor_y-1.)
        if isinstance(self, BoundsI):
            dx = int(math.ceil(dx))
            dy = int(math.ceil(dy))
        return self.withBorder(dx,dy)

    def isDefined(self):
        "Test whether `Bounds` rectangle is defined."
        return self._isdefined

    def getXMin(self):
        "Get the value of xmin."
        return self.xmin

    def getXMax(self):
        "Get the value of xmax."
        return self.xmax

    def getYMin(self):
        "Get the value of ymin."
        return self.ymin

    def getYMax(self):
        "Get the value of ymax."
        return self.ymax

    def shift(self, delta):
        """Shift the `Bounds` instance by a supplied `Position`.

        Examples:

        The shift method takes either a `PositionI` or `PositionD` instance, which must match
        the type of the `Bounds` instance::

            >>> bounds = BoundsI(1,32,1,32)
            >>> bounds = bounds.shift(galsim.PositionI(3, 2))
            >>> bounds = BoundsD(0, 37.4, 0, 49.9)
            >>> bounds = bounds.shift(galsim.PositionD(3.9, 2.1))
        """
        if not isinstance(delta, self._pos_class):
            raise TypeError("delta must be a %s instance"%self._pos_class)
        return self.__class__(self.xmin + delta.x, self.xmax + delta.x,
                              self.ymin + delta.y, self.ymax + delta.y)

    def __and__(self, other):
        if not isinstance(other, self.__class__):
            raise TypeError("other must be a %s instance"%self.__class__.__name__)
        if not self.isDefined() or not other.isDefined():
            return self.__class__()
        else:
            xmin = max(self.xmin, other.xmin)
            xmax = min(self.xmax, other.xmax)
            ymin = max(self.ymin, other.ymin)
            ymax = min(self.ymax, other.ymax)
            if xmin > xmax or ymin > ymax:
                return self.__class__()
            else:
                return self.__class__(xmin, xmax, ymin, ymax)

    def __add__(self, other):
        if isinstance(other, self.__class__):
            if not other.isDefined():
                return self
            elif self.isDefined():
                xmin = min(self.xmin, other.xmin)
                xmax = max(self.xmax, other.xmax)
                ymin = min(self.ymin, other.ymin)
                ymax = max(self.ymax, other.ymax)
                return self.__class__(xmin, xmax, ymin, ymax)
            else:
                return other
        elif isinstance(other, self._pos_class):
            if self.isDefined():
                xmin = min(self.xmin, other.x)
                xmax = max(self.xmax, other.x)
                ymin = min(self.ymin, other.y)
                ymax = max(self.ymax, other.y)
                return self.__class__(xmin, xmax, ymin, ymax)
            else:
                return self.__class__(other)
        else:
            raise TypeError("other must be either a %s or a %s"%(
                            self.__class__.__name__, self._pos_class.__name__))

    def __repr__(self):
        if self.isDefined():
            return "galsim.%s(xmin=%r, xmax=%r, ymin=%r, ymax=%r)"%(
                self.__class__.__name__, self.xmin, self.xmax, self.ymin, self.ymax)
        else:
            return "galsim.%s()"%(self.__class__.__name__)

    def __str__(self):
        if self.isDefined():
            return "galsim.%s(%s,%s,%s,%s)"%(
                self.__class__.__name__, self.xmin, self.xmax, self.ymin, self.ymax)
        else:
            return "galsim.%s()"%(self.__class__.__name__)

    def _getinitargs(self):
        if self.isDefined():
            return (self.xmin, self.xmax, self.ymin, self.ymax)
        else:
            return ()

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, self.__class__) and self._getinitargs() == other._getinitargs()))

    def __ne__(self, other):
        return not self.__eq__(other)

    def __hash__(self):
        return hash((self.__class__.__name__, self._getinitargs()))


class BoundsD(Bounds):
    """A `Bounds` that takes floating point values.

    See the `Bounds` doc string for more details.
    """
    _pos_class = PositionD

    def __init__(self, *args, **kwargs):
        self._parse_args(*args, **kwargs)
        self.xmin = float(self.xmin)
        self.xmax = float(self.xmax)
        self.ymin = float(self.ymin)
        self.ymax = float(self.ymax)

    @property
    def _b(self):
        return _galsim.BoundsD(float(self.xmin), float(self.xmax),
                               float(self.ymin), float(self.ymax))

    def _check_scalar(self, x, name):
        try:
            if x == float(x): return
        except (TypeError, ValueError):
            pass
        raise TypeError("%s must be a float value"%name)

    def _area(self):
        return (self.xmax - self.xmin) * (self.ymax - self.ymin)

    @property
    def _center(self):
        return _PositionD( (self.xmax + self.xmin)/2., (self.ymax + self.ymin)/2. )


class BoundsI(Bounds):
    """A `Bounds` that takes only integer values.

    Typically used to define the bounding box of an image.

    See the `Bounds` doc string for more details.
    """
    _pos_class = PositionI

    def __init__(self, *args, **kwargs):
        self._parse_args(*args, **kwargs)
        if (self.xmin != int(self.xmin) or self.xmax != int(self.xmax) or
            self.ymin != int(self.ymin) or self.ymax != int(self.ymax)):
            raise TypeError("BoundsI must be initialized with integer values")
        # Now make sure they are all ints
        self.xmin = int(self.xmin)
        self.xmax = int(self.xmax)
        self.ymin = int(self.ymin)
        self.ymax = int(self.ymax)

    @property
    def _b(self):
        return _galsim.BoundsI(self.xmin, self.xmax, self.ymin, self.ymax)

    def _check_scalar(self, x, name):
        try:
            if x == int(x): return
        except (TypeError, ValueError):
            pass
        raise TypeError("%s must be an integer value"%name)

    def numpyShape(self):
        "A simple utility function to get the numpy shape that corresponds to this `Bounds` object."
        if self.isDefined():
            return self.ymax-self.ymin+1, self.xmax-self.xmin+1
        else:
            return 0,0

    def _area(self):
        # Remember the + 1 this time to include the pixels on both edges of the bounds.
        if not self.isDefined():
            return 0
        else:
            return (self.xmax - self.xmin + 1) * (self.ymax - self.ymin + 1)

    @property
    def _center(self):
        # Write it this way to make sure the integer rounding goes the same way regardless
        # of whether the values are positive or negative.
        # e.g. (1,10,1,10) -> (6,6)
        #      (-10,-1,-10,-1) -> (-5,-5)
        # Just up and to the right of the true center in both cases.
        return _PositionI(self.xmin + (self.xmax - self.xmin + 1)//2,
                          self.ymin + (self.ymax - self.ymin + 1)//2)


def _BoundsD(xmin, xmax, ymin, ymax):
    """Equivalent to `BoundsD` constructor, but skips some sanity checks and argument parsing.
    This requires that the four values be float types.
    """
    ret = BoundsD.__new__(BoundsD)
    ret._isdefined = True
    ret.xmin = float(xmin)
    ret.xmax = float(xmax)
    ret.ymin = float(ymin)
    ret.ymax = float(ymax)
    return ret


def _BoundsI(xmin, xmax, ymin, ymax):
    """Equivalent to `BoundsI` constructor, but skips some sanity checks and argument parsing.
    This requires that the four values be int types.
    """
    ret = BoundsI.__new__(BoundsI)
    ret._isdefined = True
    ret.xmin = int(xmin)
    ret.xmax = int(xmax)
    ret.ymin = int(ymin)
    ret.ymax = int(ymax)
    return ret




# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#

__all__ = [ 'VonKarman' ]

import numpy as np
import astropy.units as u

from . import _galsim
from .gsobject import GSObject
from .gsparams import GSParams
from .utilities import lazy_property, doc_inherit
from .position import _PositionD
from .angle import arcsec, AngleUnit
from .errors import GalSimError, convert_cpp_errors, galsim_warn
from .errors import GalSimIncompatibleValuesError


class VonKarman(GSObject):
    r"""Class describing a von Karman surface brightness profile, which represents a long exposure
    atmospheric PSF.  The difference between the von Karman profile and the related `Kolmogorov`
    profile is that the von Karman profile includes a parameter for the outer scale of atmospheric
    turbulence, which is a physical scale beyond which fluctuations in the refractive index stop
    growing, typically between 10 and 100 meters.  Quantitatively, the von Karman phase fluctuation
    power spectrum at spatial frequency f is proportional to

    .. math::

        P(f) = r_0^{-5/3} \left(f^2 + L_0^{-2}\right)^{-11/6}

    where :math:`r_0` is the Fried parameter which sets the overall turbulence amplitude and
    :math:`L_0` is the outer scale in meters.
    The Kolmogorov power spectrum proportional to :math:`r_0^{-5/3} f^{-11/3}` is recovered
    as :math:`L_0 \rightarrow \infty`.

    For more information, we recommend the following references:

        Martinez et al.  2010  A&A  vol. 516
        Conan  2008  JOSA  vol. 25

    .. note::

        If one blindly follows the math for converting the von Karman power spectrum into a PSF, one
        finds that the PSF contains a delta-function at the origin with fractional flux of:

        .. math::

            F_\mathrm{delta} = e^{-0.086 (r_0/L_0)^{-5/3}}

        In almost all cases of interest this evaluates to something tiny, often on the order of
        :math:`10^{-100}` or smaller.  By default, GalSim will ignore this delta function entirely
        since it usually doesn't make any difference, but can complicate some calculations like
        drawing using method='real_space' or by formally requiring huge Fourier transforms for
        drawing using method='fft'.  If for some reason you want to keep the delta function
        though, then you can pass the do_delta=True argument to the VonKarman initializer.

    Parameters:
        lam:                Wavelength, either as an astropy Quantity or a float in nanometers.
        r0:                 Fried parameter at specified wavelength ``lam``, either as an astropy
                            Quantity or a float in meters. Exactly one of r0 and r0_500 should be
                            specified.
        r0_500:             Fried parameter at 500 nm, either as an astropy Quantity or a float in
                            meters. Exactly one of r0 and r0_500 should be specified.
        L0:                 Outer scale, either as an astropy Quantity or a float in meters.
                            [default: 25.0 m]
        flux:               The flux (in photons/cm^2/s) of the profile. [default: 1]
        scale_unit:         Units assumed when drawing this profile or evaluating xValue, kValue,
                            etc.  Should be a `galsim.AngleUnit` or a string that can be used to
                            construct one (e.g., 'arcsec', 'radians', etc.).
                            [default: galsim.arcsec]
        force_stepk:        By default, VonKarman will derive a value of stepk from a computed
                            real-space surface brightness profile and gsparams settings.  Although
                            this profile is cached for future instantiations of identical VonKarman
                            objects, it is relatively slow to compute for the first instance and
                            can dominate the compute time when drawing many VonKaman's with
                            different parameters using method 'fft', 'sb', or 'no_pixel', a
                            situation that may occur, e.g., in a fitting context.  This keyword
                            enables a user to bypass the real-space profile computation by directly
                            specifying a stepk value.  Note that if the ``.half_light_radius``
                            property is queried, or the object is drawn using method 'phot' or
                            'real_space', then the real-space profile calculation is performed (if
                            not cached) at that point.  [default: 0.0, which means do not force a
                            value of stepk]
        do_delta:           Include delta-function at origin? (not recommended; see above).
                            [default: False]
        suppress_warning:   For some combinations of r0 and L0, it may become impossible to satisfy
                            the gsparams.maxk_threshold criterion (see above).  In that case, the
                            code will emit a warning alerting the user that they may have entered a
                            non-physical regime.  However, this warning can be suppressed with this
                            keyword.  [default: False]
        gsparams:           An optional `GSParams` argument. [default: None]
    """
    _req_params = { "lam" : (float, u.Quantity) }
    _opt_params = {
        "L0" : (float, u.Quantity),
        "flux" : float,
        "scale_unit" : str,
        "do_delta" : bool
    }
    _single_params = [ {
        "r0" : (float, u.Quantity),
        "r0_500" : (float, u.Quantity)
    } ]

    _has_hard_edges = False
    _is_axisymmetric = True
    #_is_analytic_x = True  # = not do_delta  defined below.
    _is_analytic_k = True

    def __init__(self, lam, r0=None, r0_500=None, L0=25.0, flux=1, scale_unit=arcsec,
                 force_stepk=0.0, do_delta=False, suppress_warning=False, gsparams=None):
        if isinstance(lam, u.Quantity):
            lam = lam.to_value(u.nm)
        if isinstance(r0, u.Quantity):
            r0 = r0.to_value(u.m)
        if isinstance(r0_500, u.Quantity):
            r0_500 = r0_500.to_value(u.m)
        if isinstance(L0, u.Quantity):
            L0 = L0.to_value(u.m)

        # We lose stability if L0 gets too large.  This should be close enough to infinity for
        # all practical purposes though.
        if L0 > 1e10:
            L0 = 1e10

        if r0 is not None and r0_500 is not None:
            raise GalSimIncompatibleValuesError(
                "Only one of r0 and r0_500 may be specified",
                r0=r0, r0_500=r0_500)
        if r0 is None and r0_500 is None:
            raise GalSimIncompatibleValuesError(
                "Either r0 or r0_500 must be specified",
                r0=r0, r0_500=r0_500)
        if r0_500 is not None:
            r0 = r0_500 * (lam/500.)**1.2

        if isinstance(scale_unit, str):
            self._scale_unit = AngleUnit.from_name(scale_unit)
        else:
            self._scale_unit = scale_unit
        self._scale = self._scale_unit/arcsec

        self._lam = float(lam)
        self._r0 = float(r0)
        self._L0 = float(L0)
        self._flux = float(flux)
        self._do_delta = bool(do_delta)
        self._gsparams = GSParams.check(gsparams)
        self._suppress = bool(suppress_warning)
        self._force_stepk = force_stepk
        self._sbvk  # Make this now, so we get the warning if appropriate.

    @lazy_property
    def _sbvk(self):
        with convert_cpp_errors():
            sbvk = _galsim.SBVonKarman(self._lam, self._r0, self._L0, self._flux,
                                       self._scale, self._do_delta, self._gsparams._gsp,
                                       self._force_stepk)

        self._delta = sbvk.getDelta()
        if not self._suppress:
            if self._delta > self._gsparams.maxk_threshold:
                galsim_warn("VonKarman delta-function component is larger than maxk_threshold.  "
                            "Please see docstring for information about this component and how "
                            "to toggle it.")
        if self._do_delta:
            with convert_cpp_errors():
                sbvk = _galsim.SBVonKarman(self._lam, self._r0, self._L0,
                                           self._flux-self._delta, self._scale,
                                           self._do_delta, self._gsparams._gsp,
                                           self._force_stepk)
        return sbvk

    @lazy_property
    def _sbp(self):
        # Add in a delta function with appropriate amplitude if requested.
        if self._do_delta:
            sbvk = self._sbvk
            sbdelta = _galsim.SBDeltaFunction(self._delta, self._gsparams._gsp)
            return _galsim.SBAdd([sbvk, sbdelta], self._gsparams._gsp)
        else:
            return self._sbvk

    @property
    def lam(self):
        """The input lam value.
        """
        return self._lam

    @property
    def r0(self):
        """The input r0 value.
        """
        return self._r0

    @property
    def r0_500(self):
        """The input r0_500 value.
        """
        return self._r0*(self._lam/500.)**(-1.2)

    @property
    def L0(self):
        """The input L0 value.
        """
        return self._L0

    @property
    def scale_unit(self):
        """The input scale_units.
        """
        return self._scale_unit

    @property
    def force_stepk(self):
        """Forced value of stepk or 0.0.
        """
        return self._force_stepk

    @property
    def do_delta(self):
        """Whether to include the delta function at the center.
        """
        return self._do_delta

    @property
    def _is_analytic_x(self):
        return not self._do_delta

    @property
    def delta_amplitude(self):
        """The amplitude of the delta function at the center.
        """
        self._sbvk  # This is where _delta is calculated.
        return self._delta

    @property
    def half_light_radius(self):
        """The half-light radius.
        """
        return self._sbvk.getHalfLightRadius()

    def _structure_function(self, rho):
        return self._sbvk.structureFunction(rho)

    def __eq__(self, other):
        return (self is other or
                (isinstance(other, VonKarman) and
                 self.lam == other.lam and
                 self.r0 == other.r0 and
                 self.L0 == other.L0 and
                 self.flux == other.flux and
                 self.scale_unit == other.scale_unit and
                 self.force_stepk == other.force_stepk and
                 self.do_delta == other.do_delta and
                 self.gsparams == other.gsparams))

    def __hash__(self):
        return hash(("galsim.VonKarman", self.lam, self.r0, self.L0, self.flux, self.scale_unit,
                     self.force_stepk, self.do_delta, self.gsparams))

    def __repr__(self):
        out = "galsim.VonKarman(lam=%r, r0=%r, L0=%r"%(self.lam, self.r0, self.L0)
        out += ", flux=%r"%self.flux
        if self.scale_unit != arcsec:
            out += ", scale_unit=%r"%self.scale_unit
        if self.force_stepk:
            out += ", force_stepk=%r"%self.force_stepk
        if self.do_delta:
            out += ", do_delta=True"
        if self._suppress:
            out += ", suppress_warning=True"
        out += ", gsparams=%r"%self.gsparams
        out += ")"
        return out

    def __str__(self):
        return "galsim.VonKarman(lam=%r, r0=%r, L0=%r)"%(self.lam, self.r0, self.L0)

    def __getstate__(self):
        d = self.__dict__.copy()
        d.pop('_sbp',None)
        d.pop('_sbvk',None)
        return d

    def __setstate__(self, d):
        self.__dict__ = d

    @property
    def _maxk(self):
        return self._sbvk.maxK()

    @property
    def _stepk(self):
        return self._sbvk.stepK()

    @property
    def _max_sb(self):
        return self._sbvk.xValue(_PositionD(0,0)._p)

    def _xValue(self, pos):
        return self._sbvk.xValue(pos._p)

    def _kValue(self, kpos):
        return self._sbp.kValue(kpos._p)

    def _drawReal(self, image, jac=None, offset=(0.,0.), flux_scaling=1.):
        _jac = 0 if jac is None else jac.__array_interface__['data'][0]
        dx,dy = offset
        self._sbp.draw(image._image, image.scale, _jac, dx, dy, flux_scaling)

    def _shoot(self, photons, rng):
        self._sbp.shoot(photons._pa, rng._rng)

    def _drawKImage(self, image, jac=None):
        _jac = 0 if jac is None else jac.__array_interface__['data'][0]
        self._sbp.drawK(image._image, image.scale, _jac)

    @doc_inherit
    def withFlux(self, flux):
        return VonKarman(lam=self.lam, r0=self.r0, L0=self.L0, flux=flux,
                         scale_unit=self.scale_unit,
                         force_stepk=self.force_stepk, do_delta=self.do_delta,
                         suppress_warning=self._suppress, gsparams=self.gsparams)



# Copyright (c) 2012-2026 by the GalSim developers team on GitHub
# https://github.com/GalSim-developers
#
# This file is part of GalSim: The modular galaxy image simulation toolkit.
# https://github.com/GalSim-developers/GalSim
#
# GalSim is free software: redistribution and use in source and binary forms,
# with or without modification, are permitted provided that the following
# conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions, and the disclaimer given in the accompanying LICENSE
#    file.
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions, and the disclaimer given in the documentation
#    and/or other materials provided with the distribution.
#
"""
@file roman_backgrounds.py

Part of the Roman Space Telescope module.  This file includes any routines needed to define the
background level, for which the main contribution is zodiacal light.
"""

import numpy as np
import os

from .. import degrees, radians, CelestialCoord
from .. import GalSimValueError

def getSkyLevel(bandpass, world_pos=None, exptime=None, epoch=2025, date=None):
    """
    Get the expected sky level for a Roman ST observation due to zodiacal light for this bandpass
    and position.

    This routine requires Bandpass objects that were loaded by galsim.roman.getBandpasses().  That
    routine will have stored tables containing the sky background as a function of position on the
    sky for that bandpass.  This routine then interpolates between the values in those tables to
    arbitrary positions on the sky.

    The numbers that are stored in the Bandpass object ``bandpass`` are background level in units of
    e-/s/arcsec^2.  Multiplying by the exposure time gives a result in e-/arcsec^2.  The
    result can either be multiplied by the approximate pixel area to get e-/pix, or the result can
    be used with wcs.makeSkyImage() to make an image of the sky that properly includes the actual
    pixel area as a function of position on the detector.

    The source of the tables that are being interpolated is Chris Hirata's publicly-available Roman
    exposure time calculator (ETC):

        http://www.tapir.caltech.edu/~chirata/web/software/space-etc/

    Using the throughput files loaded into the ``bandpass`` object, the calculation nominally
    returns photons/s/arcsec^2, but the input bandpasses used internally by the ETC
    code include the quantum efficiency, to effectively convert to e-/s/arcsec^2.  Note that in
    general results will depend on the adopted model for zodiacal light, and these are uncertain at
    the ~10% level. One must also better sample the integration in the zodiacal light calculation
    to match the output tables used by GalSim here.

    Positions should be specified with the ``world_pos`` keyword, which must be a CelestialCoord
    object.  If no ``world_pos`` is supplied, then the routine will use a default position that
    looks sensibly away from the sun.

    Parameters:
        bandpass:   A Bandpass object.
        world_pos:  A position, given as a CelestialCoord object.  If None, then the routine
                    will use an ecliptic longitude of 90 degrees with respect to the sun
                    position (as a fair compromise between 0 and 180), and an ecliptic latitude
                    of 30 degrees with respect to the sun position (decently out of the plane
                    of the Earth-sun orbit). [default: None]
        exptime:    Exposure time in seconds.  If None, use the default Roman exposure time.
                    [default: None]
        epoch:      The epoch to be used for estimating the obliquity of the ecliptic when
                    converting ``world_pos`` to ecliptic coordinates.  This keyword is only used
                    if ``date`` is None, otherwise ``date`` is used to determine the ``epoch``.
                    [default: 2025]
        date:       The date of the observation, provided as a python datetime object.  If None,
                    then the conversion to ecliptic coordinates assumes the sun is at ecliptic
                    coordinates of (0,0), as it is at the vernal equinox. [default: None]

    Returns:
        the expected sky level in e-/arcsec^2.
    """
    if exptime is None:
        from . import exptime

    # Check for cached sky level information for this filter.  If not, raise exception
    if not hasattr(bandpass, '_sky_level'):
        raise GalSimValueError("Only bandpasses returned from galsim.roman.getBandpasses() are "
                               "allowed here!", bandpass)

    # Check for proper type for position, and extract the ecliptic coordinates.
    if world_pos is None:
        # Use our defaults for the case of unspecified position.
        ecliptic_lat = 30.*degrees
        ecliptic_lon = 90.*degrees
    else:
        if not isinstance(world_pos, CelestialCoord):
            raise TypeError("world_pos must be supplied as a CelestialCoord.")
        if date is not None:
            epoch = date.year
        ecliptic_lon, ecliptic_lat = world_pos.ecliptic(epoch=epoch, date=date)

    # Check the position in our table, and make sure to take advantage of the latitude / longitude
    # symmetries:
    # The table only includes positive values of latitude, because there is symmetry about zero.  So
    # we take the absolute value of the input ecliptic latitude.
    # The table only includes longitude in the range [0, 180] because there is symmetry in that a
    # negative longitude in the range[-180, 0] should have the same sky level as at the positive
    # value of longitude (given that the Sun is at 0).
    ecliptic_lon = ecliptic_lon.wrap()
    ecliptic_lon = abs(ecliptic_lon.rad)*radians
    ecliptic_lat = abs(ecliptic_lat.rad)*radians
    sin_ecliptic_lat = np.sin(ecliptic_lat)

    # Take the lookup table, and turn negative numbers (indicating failure because of proximity to
    # sun) to large positive values so that we can identify them as bad after interpolation.
    max_sky = np.max(bandpass._sky_level)
    sky_level = bandpass._sky_level.copy()
    sky_level[sky_level<0] = 1.e6

    # Interpolate in 2d on the table.
    s = sky_level.reshape(46,42).transpose()
    xlat = sin_ecliptic_lat*41
    xlon = abs(ecliptic_lon.wrap() / degrees)/4.
    ilat = int(xlat)
    ilon = int(xlon)
    xlat -= ilat
    xlon -= ilon
    sky_val = (s[ilat, ilon] * (1.-xlat)*(1.-xlon) +
               s[ilat, ilon+1] * (1.-xlat)*xlon +
               s[ilat+1, ilon] * xlat*(1.-xlon) +
               s[ilat+1, ilon+1] * xlat*xlon)

    # If the result is too large, then raise an exception: we should not look at this position!
    if sky_val > max_sky:
        raise GalSimValueError("world_pos is too close to sun. Would not observe here.", world_pos)

    # Now, convert to the right units, and return.  (See docstring for explanation.)
    # Multiply by exposure time.
    sky_val *= exptime

    # The result is now the sky level in e-/arcsec^2.
    return sky_val
